"""This Module allows for the listing of users passwords"""
from libpkpass.util import dictionary_filter
from libpkpass.commands.command import Command
from libpkpass.errors import CliArgumentError


class List(Command):
    ####################################################################
    """This class implements the cli list"""
    ####################################################################
    name = "list"
    description = "List passwords you have access to"
    selected_args = Command.selected_args + ["pwstore", "stdin", "recovery", "filter"]

    def _run_command_execution(self):
        ####################################################################
        """Run function for class."""
        ####################################################################
        result = {}
        for pwname, passwordentry in self.passworddb.pwdb.items():
            if self.args["recovery"] and passwordentry.escrow:
                recipients = passwordentry.escrow["recipients"]
                for key, value in recipients.items():
                    if key == self.args["identity"]:
                        result[pwname] = {
                            "name": passwordentry.metadata["name"],
                            "stake_holders": list(recipients.keys()),
                            "distributor": value["distributor"],
                            "minimum_shares": passwordentry.escrow["metadata"][
                                "minimum_escrow"
                            ],
                        }
            elif (
                not self.args["recovery"]
                and self.args["identity"] in passwordentry.recipients.keys()
            ):
                result[pwname] = {
                    "name": passwordentry.metadata["name"],
                    "distributor": passwordentry.recipients[self.args["identity"]][
                        "distributor"
                    ],
                }

        if "filter" in self.args and self.args["filter"]:
            result = dictionary_filter(self.args["filter"], result)

        yield f"Passwords for '{self.args['identity']}':"
        for key, value in sorted(result.items()):
            yield f"{self.color_print(key + ':', 'first_level')}\n  {self.color_print('Distributor: ', 'second_level') + value['distributor']}\n  {self.color_print('Name: ', 'second_level') + value['name']}"

    def _validate_args(self):
        for argument in ["keypath"]:
            if argument not in self.args or self.args[argument] is None:
                raise CliArgumentError(f"'{argument}' is a required argument")



from sqlalchemy import String, Boolean, DateTime
from sqlalchemy.schema import Column
from libpkpass.models import Base


class Cert(Base):
    #########################################
    """Table of Certificates"""
    #########################################

    __tablename__ = "cert"
    cert_bytes = Column(String())
    verified = Column(Boolean)
    fingerprint = Column(String(), primary_key=True)
    subject = Column(String())
    issuer = Column(String())
    enddate = Column(DateTime())
    issuerhash = Column(String())
    subjecthash = Column(String())

    def __repr__(self):
        return f"<Cert {self.fingerprint}>"

    def __iter__(self):
        for col in self.__table__.columns:
            yield col.name, getattr(self, col.name)



"""This Module defines the arguments that argparse will accept from the CLI"""
from json import loads

ARGUMENTS = {
    ############################################################################
    # Data structure containing all arguments a user could pass into pkpass
    # This helps us de-dupe our work and keep args consistent between
    # subcommands.  We use this with *args and **kwargs constructs
    ############################################################################
    "all": {
        "args": ["-a", "--all"],
        "kwargs": {
            "help": "Show all available password to the given user, if a pwname is supplied filtering will be done case-insensitivey based on the filename",
            "action": "store_true",
        },
    },
    "authorizer": {
        "args": ["--authorizer"],
        "kwargs": {
            "help": "The person or account authorizing the creation of this secret",
            "type": str,
        },
    },
    "behalf": {
        "args": ["-b", "--behalf"],
        "kwargs": {
            "help": "Show passwords for a user using a password as its private key",
            "type": str,
        },
    },
    "cabundle": {
        "args": ["--cabundle"],
        "kwargs": {"type": str, "help": "Path to CA certificate bundle file"},
    },
    "card_slot": {
        "args": ["-c", "--card_slot"],
        "kwargs": {
            "type": str,
            "help": "The slot number of the card that should be used",
        },
    },
    "certpath": {
        "args": ["--certpath"],
        "kwargs": {
            "type": str,
            "help": "Path to directory containing public keys.  Certificates must end in '.cert'",
        },
    },
    "color": {
        "args": ["--color"],
        "kwargs": {"type": str, "help": "Disable color or not, accepts true/false"},
    },
    "connect": {
        "args": ["--connect"],
        "kwargs": {
            "type": loads,
            "help": "Connection string for the api to retrieve certs",
        },
    },
    "description": {
        "args": ["-d", "--description"],
        "kwargs": {"help": "A description of this secret", "type": str},
    },
    "escrow_users": {
        "args": ["-e", "--escrow_users"],
        "kwargs": {
            "help": "Escrow users list is a comma sepearated list of recovery users that each get part of a key",
            "type": str,
        },
    },
    "filter": {
        "args": ["-f", "--filter"],
        "kwargs": {"help": "Reduce output of commands to matching items", "type": str},
    },
    "groups": {
        "args": ["-g", "--groups"],
        "kwargs": {"type": str, "help": "Comma seperated list of recipient groups"},
    },
    "identity": {
        "args": ["-i", "--identity"],
        "kwargs": {
            "type": str,
            "help": "Override identity of user running the program",
        },
    },
    "ignore_decrypt": {
        "args": ["-I", "--ignore-decrypt"],
        "kwargs": {
            "action": "store_true",
            "help": "Ignore decryption errors during show all process",
        },
    },
    "keypath": {
        "args": ["--keypath"],
        "kwargs": {
            "type": str,
            "help": "Path to directory containing private keys.  Keys must end in '.key'",
        },
    },
    "long_escrow": {
        "args": ["-l", "--long-escrow"],
        "kwargs": {
            "action": "store_true",
            "help": "Long list passwords that have been escrowed, only shows escrowed passwords",
        },
    },
    "min_escrow": {
        "args": ["-m", "--min_escrow"],
        "kwargs": {
            "type": int,
            "help": "Minimum number of users required to unlock escrowed password",
        },
    },
    "nocache": {
        "args": ["--no-cache"],
        "kwargs": {
            "action": "store_true",
            "help": "if using a connector, pull the certs again",
        },
    },
    "noescrow": {
        "args": ["--noescrow"],
        "kwargs": {
            "action": "store_true",
            "help": "Do not use escrow functionality, ignore defaults in rc file",
        },
    },
    "nocrypto": {
        "args": ["--nocrypto"],
        "kwargs": {
            "action": "store_true",
            "help": "Do not use a password for import/export files",
        },
    },
    "nopassphrase": {
        "args": ["--nopassphrase", "--nopin"],
        "kwargs": {
            "action": "store_true",
            "help": "Do not prompt for a pin/passphrase",
        },
    },
    "nosign": {
        "args": ["--nosign"],
        "kwargs": {
            "action": "store_true",
            "help": "Do not digitally sign the password information that you are generating",
        },
    },
    "noverify": {
        "args": ["--noverify"],
        "kwargs": {
            "action": "store_true",
            "help": "Do not verify certificates and signatures",
        },
    },
    "overwrite": {
        "args": ["--overwrite"],
        "kwargs": {
            "action": "store_true",
            "help": "Overwrite a password that already exists",
        },
    },
    "pwfile": {
        "args": ["pwfile"],
        "kwargs": {
            "type": str,
            "help": "path to the import/export file",
            "nargs": "?",
            "default": None,
        },
    },
    "pwname": {
        "args": ["pwname"],
        "kwargs": {
            "type": str,
            "help": "Name of the password. Ex: passwords/team/infrastructure/root",
            "nargs": "?",
            "default": None,
        },
    },
    "pwstore": {
        "args": ["--pwstore", "--srcpwstore"],
        "kwargs": {
            "type": str,
            "help": 'Path to the source password store.  Defaults to "./passwords"',
        },
    },
    "quiet": {
        "args": ["-q", "--quiet"],
        "kwargs": {
            "action": "store_const",
            "const": -1,
            "default": 0,
            "dest": "verbosity",
            "help": "quiet output (show errors only)",
        },
    },
    "recovery": {
        "args": ["-r", "--recovery"],
        "kwargs": {
            "action": "store_true",
            "help": "Work with passwords distributed through escrow functionality",
        },
    },
    "rename": {
        "args": ["rename"],
        "kwargs": {
            "type": str,
            "help": "New name of the password.",
            "nargs": "?",
            "default": None,
        },
    },
    "rules": {
        "args": ["-R", "--rules"],
        "kwargs": {"type": str, "help": "Key of rules to use from provided rules map"},
    },
    "rules_map": {
        "args": ["--rules-map"],
        "kwargs": {
            "type": loads,
            "help": "Map of rules used for automated generation of passwords",
        },
    },
    "stdin": {
        "args": ["--stdin"],
        "kwargs": {
            "action": "store_true",
            "help": "Take all password input from stdin instead of from a user input prompt",
        },
    },
    "time": {
        "args": ["-t", "--time"],
        "kwargs": {
            "type": int,
            "help": "Number of seconds to keep password in paste buffer",
        },
    },
    "theme_map": {
        "args": ["--theme-map"],
        "kwargs": {"type": loads, "help": "Map of colors to use for colorized output"},
    },
    "type": {
        "args": ["--type"],
        "kwargs": {"type": str, "help": "Type of password integration used"},
    },
    "users": {
        "args": ["-u", "--users"],
        "kwargs": {"type": str, "help": "Comma seperated list of recipients"},
    },
    "value": {
        "args": ["--value"],
        "kwargs": {
            "action": "store_true",
            "help": "Don't update files directly, just dump value onto screen",
        },
    },
    "verbosity": {
        "args": ["-v", "--verbose"],
        "kwargs": {
            "action": "count",
            "dest": "verbosity",
            "default": 0,
            "help": "verbose output (repeat for increased verbosity)",
        },
    },
    "SCBackend": {
        "args": ["--scbackend"],
        "kwargs": {
            "type": str,
            "default": "opensc",
            "help": "SC backend to use: opensc or yubi",
        },
    },
    "PKCS11_module_path": {
        "args": ["--PKCS11-module-path"],
        "kwargs": {
            "type": str,
            "default": "/usr/local/lib/libykcs11.dylib",
            "help": "Path to yubi PKCS11 module",
        },
    },
}



"""This module allows for the deletion of passwords"""
from os import path, remove
from sys import exit as sysexit
from libpkpass.commands.command import Command
from libpkpass.errors import CliArgumentError, NotThePasswordOwnerError, PasswordIOError


class Delete(Command):
    ####################################################################
    """This class implements the CLI functionality of deletion of passwords"""
    ####################################################################
    name = "delete"
    description = "Delete a password in the repository"
    selected_args = Command.selected_args + [
        "pwname",
        "pwstore",
        "overwrite",
        "stdin",
        "keypath",
        "card_slot",
    ]

    def _run_command_execution(self):
        ####################################################################
        """Run function for class."""
        ####################################################################
        safe, owner = self.safety_check()
        if safe or self.args["overwrite"]:
            self._confirmation()
        else:
            raise NotThePasswordOwnerError(
                self.args["identity"], owner, self.args["pwname"]
            )
        # necessary for print statement
        yield ""

    def delete_pass(self):
        ##################################################################
        """This deletes a password that the user has created, useful for testing"""
        ##################################################################
        filepath = path.join(self.args["pwstore"], self.args["pwname"])
        try:
            remove(filepath)
        except OSError as err:
            raise PasswordIOError(
                f"Password '{self.args['pwname']}' not found"
            ) from err

    def _confirmation(self):
        ####################################################################
        """Verify password to modify"""
        ####################################################################
        yes = {"yes", "y", "ye", ""}
        deny = {"no", "n"}
        confirmation = input(
            f"{self.args['pwname']}: \nDelete this password?(Defaults yes): "
        )
        if confirmation.lower() in yes:
            self.delete_pass()
        elif confirmation.lower() in deny:
            sysexit()
        else:
            print("please respond with yes or no")
            self._confirmation()

    def _validate_args(self):
        for argument in ["pwname", "keypath"]:
            if argument not in self.args or self.args[argument] is None:
                raise CliArgumentError(f"'{argument}' is a required argument")



"""This Module allows for the import of passwords"""
import getpass
import os.path
from yaml import safe_load, scanner
from tqdm import tqdm
from libpkpass import LOGGER
from libpkpass.crypto import sk_decrypt_string
from libpkpass.commands.command import Command
from libpkpass.password import PasswordEntry
from libpkpass.errors import CliArgumentError, LegacyImportFormatError, FileOpenError


class Import(Command):
    ####################################################################
    """This Class implements the cli functionality of import"""
    ####################################################################
    name = "import"
    description = "Import passwords that you have saved to a file"
    selected_args = Command.selected_args + [
        "pwfile",
        "stdin",
        "nopassphrase",
        "card_slot",
        "nocrypto",
    ]

    def _run_command_execution(self):
        ####################################################################
        """Run function for class."""
        ####################################################################
        try:
            contents = ""
            with open(self.args["pwfile"], "r", encoding="ASCII") as fcontents:
                contents = fcontents.read().strip()
            if self.args["nocrypto"]:
                self._file_handler(contents)
            else:
                passwd = getpass.getpass("Please enter the password for the file: ")
                passwords = contents.split("\n")
                for password in tqdm(passwords):
                    self._file_handler(sk_decrypt_string(password, passwd))
        except IOError as err:
            raise FileOpenError(
                self.args["pwfile"], "No such file or directory"
            ) from err
        # necessary for print statement
        yield ""

    def _file_handler(self, string):
        ####################################################################
        """This function handles the contents of a file"""
        ####################################################################
        try:
            self._yaml_file(safe_load(string))
        except (TypeError, scanner.ScannerError, KeyError):
            try:
                self._flat_file(string.strip().split("\n"))
            except TypeError as err:
                raise LegacyImportFormatError from err

    def _flat_file(self, passwords):
        ####################################################################
        """This function handles the simple key:value pair"""
        ####################################################################
        LOGGER.info(
            "Flat password file detected, using 'imported' as description \
you can manually change the description in the file if you would like"
        )
        for password in tqdm(passwords):
            psplit = password.split(":")
            fname = psplit[0].strip()
            pvalue = psplit[1].strip()
            self.args["pwname"] = fname
            self.create_or_update_pass(pvalue, "imported", self.args["identity"])

    def _yaml_file(self, password):
        ####################################################################
        """This function handles the yaml format of pkpass"""
        ####################################################################
        uid = self.iddb.id["name"]
        pwstore = self.args["pwstore"]

        self.args["pwname"] = password["metadata"]["name"]
        plaintext_str = password["recipients"][uid]["encrypted_secret"]
        full_path = os.path.join(pwstore, password["metadata"]["name"])
        self.args["overwrite"] = True
        plist = list(password["recipients"])
        if os.path.isfile(full_path):
            existing_password = PasswordEntry()
            existing_password.read_password_data(full_path)
            self.args["overwrite"] = False
            password["metadata"] = existing_password["metadata"]
            plist += list(existing_password["recipients"])

        description = password["metadata"]["description"]
        authorizer = password["metadata"]["authorizer"]
        self.create_or_update_pass(plaintext_str, description, authorizer, plist)

    def _validate_args(self):
        ####################################################################
        """Ensure arguments are appropriate for this command"""
        ####################################################################
        for argument in ["pwfile", "keypath"]:
            if argument not in self.args or self.args[argument] is None:
                raise CliArgumentError(f"'{argument}' is a required argument")



#!/usr/bin/env python3
"""This module tests the info module"""
import unittest
import pylibyaml  # pylint: disable=unused-import
import yaml
from libpkpass.commands.cli import Cli
from libpkpass.errors import CliArgumentError
from .basetest.basetest import patch_args, ERROR_MSGS


class InfoTests(unittest.TestCase):
    """This class tests the info class"""

    def setUp(self):
        self.check_dict = {
            "Metadata": {
                "Authorizer": "y",
                "Creator": "r1",
                "Description": "y",
                "Name": "test",
                "Schemaversion": "v2",
                "Signature": "None",
            },
            "Escrow Group": {
                "Creator": "r1",
                "Share Holders": "r2, r3, r4",
                "Total Group Share Holders": 3,
                "Minimum_escrow": 2,
            },
            "Recipients": "r1",
            "Total Recipients": 1,
        }

    def test_info(self):
        """Test what info shows on a password"""
        with patch_args(
            subparser_name="info", identity="r1", nopassphrase="true", pwname="test"
        ):
            out = "\n".join(Cli().run())
        output = yaml.safe_load(out)
        del output["Earliest distribute timestamp"]
        del output["Escrow Group"]["Group creation time"]
        self.assertDictEqual(output, self.check_dict)

    def test_info_no_pass(self):
        """Test what happens when pwname is not supplied"""
        with patch_args(
            subparser_name="info", identity="r1", nopassphrase="true", pwname=None
        ):
            with self.assertRaises(CliArgumentError) as context:
                "".join(Cli().run())
        self.assertEqual(context.exception.msg, ERROR_MSGS["pwname"])


if __name__ == "__main__":
    unittest.main()



#!/usr/bin/env python3
"""This Module tests crypto functionality"""
import unittest
from sqlalchemy import create_engine
from sqlalchemy.orm import sessionmaker
from libpkpass.crypto import (
    pk_encrypt_string,
    pk_decrypt_string,
    pk_sign_string,
    pk_verify_signature,
    get_cert_fingerprint,
)
from libpkpass.identities import IdentityDB
from libpkpass.models.recipient import Recipient
from libpkpass.models.cert import Cert


class TestBasicFunction(unittest.TestCase):
    """This class tests the crypto class"""

    def setUp(self):
        self.plaintext = "PLAINTEXT"

        self.certdir = "test/pki/intermediate/certs"
        self.keydir = "test/pki/intermediate/private"
        self.cabundle = "test/pki/intermediate/certs/ca-bundle"

        self.iddb = IdentityDB()
        db_path = "test/pki/intermediate/certs/rd.db"
        self.iddb.args = {
            "db": {
                "uri": f"sqlite+pysqlite:///{db_path}",
                "engine": create_engine(f"sqlite+pysqlite:///{db_path}"),
            },
        }
        self.iddb.session = sessionmaker(bind=self.iddb.args["db"]["engine"])()
        self.iddb.load_certs_from_directory(self.certdir, self.cabundle)
        self.iddb.load_keys_from_directory(self.keydir)

    # Encrypt strings for all test identities and make sure they are different
    def test_encrypt_string(self):
        """Test encrypting a string"""
        results = []
        for identity in self.iddb.session.query(Recipient).all():
            cert = (
                self.iddb.session.query(Cert)
                .filter(Cert.recipients.contains(identity))
                .first()
            )
            results.append(pk_encrypt_string(self.plaintext, cert.cert_bytes))
        self.assertTrue(results[0] != results[1])

        # Encrypt/Decrypt strings for all test identities and make sure we get back what we put in
        for identity in self.iddb.session.query(Recipient).all():
            cert = (
                self.iddb.session.query(Cert)
                .filter(Cert.recipients.contains(identity))
                .first()
            )
            (ciphertext, derived_key) = pk_encrypt_string(
                self.plaintext, cert.cert_bytes
            )
            plaintext = pk_decrypt_string(ciphertext, derived_key, dict(identity), None)
            self.assertEqual(self.plaintext, plaintext)

    def test_verify_string(self):
        """verify string is correct"""
        results = []
        for identity in self.iddb.session.query(Recipient).all():
            results.append(pk_sign_string(self.plaintext, dict(identity), None))
        self.assertTrue(results[0] != results[1])

        for identity in self.iddb.session.query(Recipient).all():
            signature = pk_sign_string(self.plaintext, dict(identity), None)
            cert = (
                self.iddb.session.query(Cert)
                .filter(Cert.recipients.contains(identity))
                .first()
            )
            self.assertTrue(pk_verify_signature(self.plaintext, signature, [cert]))

    def test_cert_fingerprint(self):
        """Verify fingerprint is correct"""
        for identity in self.iddb.session.query(Recipient).all():
            cert = (
                self.iddb.session.query(Cert)
                .filter(Cert.recipients.contains(identity))
                .first()
            )
            fingerprint = get_cert_fingerprint(cert.cert_bytes)
            self.assertTrue(len(fingerprint.split(":")) == 20)


if __name__ == "__main__":
    unittest.main()



#!/usr/bin/env python3
"""This module tests the modify module"""
import builtins
import unittest
import mock
from libpkpass.commands.cli import Cli
from libpkpass.errors import DecryptionError
from .basetest.basetest import patch_args


class ModifyTests(unittest.TestCase):
    """This class tests the modify class"""

    def test_modify_success(self):
        """Test modifying the metadata of a password"""
        ret = True
        try:
            with patch_args(
                subparser_name="modify",
                identity="r3",
                nopassphrase="true",
                pwname="gentest",
            ):
                with mock.patch.object(builtins, "input", lambda _: "y"):
                    "".join(Cli().run())
        except DecryptionError:
            ret = False
        self.assertTrue(ret)



#!/usr/bin/env python3
"""General Utility file for common functionality"""
from os import path
from tempfile import gettempdir
from json import loads
from getpass import getuser
from sys import argv
from re import search, error
from fnmatch import filter as fnfilter
from argparse import _SubParsersAction
from sqlalchemy import create_engine
from pylibyaml import monkey_patch_pyyaml # pylint: disable=unused-import
from yaml import safe_load
from yaml.parser import ParserError
from yaml.scanner import ScannerError
from colored import fg, attr
from libpkpass import __version__
from libpkpass.models import Base
from libpkpass.models.recipient import Recipient # pylint: disable=unused-import
from libpkpass.models.cert import Cert # pylint: disable=unused-import
from libpkpass.errors import FileOpenError, JsonArgumentError, ConfigParseError, NullRecipientError, GroupDefinitionError


def color_prepare(string, color_type, colorize, theme_map=None):
    ####################################################################
    """Handle the color output of a given string"""
    ####################################################################
    if theme_map is None:
        theme_map = {}
    color_defaults = {
        "info": "cyan",
        "warning": "yellow",
        "debug": "red",
        "first_level": "magenta",
        "second_level": "green",
        "third_level": "blue",
    }
    color = (
        theme_map[color_type].lower()
        if (color_type in theme_map)
        else color_defaults[color_type]
    )
    try:
        return f"{fg(color)}{string}{attr('reset')}" if colorize else string
    except KeyError:
        return (
            f"{fg(color_defaults[color_type])}{string}{attr('reset')}"
            if colorize
            else string
        )


def set_default_subparser(self, name, args=None, positional_args=0):
    ####################################################################
    """Set default subparser to interpreter"""
    ####################################################################
    subparser_found = False
    for arg in argv[1:]:
        if arg in ["-h", "--help"]:
            break
    else:
        # pylint: disable=protected-access
        # disabling protected access because subparsers are protected
        # and we want to be able to have a default subparser
        for action in self._subparsers._actions:
            if not isinstance(action, _SubParsersAction):
                continue
            for _ in argv[1:]:
                subparser_found = True
        if not subparser_found:
            if args is None:
                argv.insert(len(argv) - positional_args, name)
            else:
                args.insert(len(args) - positional_args, name)


def show_version():
    ####################################################################
    """return the version number in the VERSION file"""
    ####################################################################
    yield __version__


def sort(lst):
    ####################################################################
    """Sort our alphanumeric keys"""
    ####################################################################
    lst = [str(i) for i in lst]
    lst.sort()
    return [int(i) if i.isdigit() else i for i in lst]


def dictionary_filter(string_match, dictionary, secondary_check=None):
    ####################################################################
    """Filter out our dictionary"""
    ####################################################################
    key_list = []
    key_list = fnfilter(dictionary.keys(), string_match)

    # If not a fileglob match, try regex
    if not key_list:
        try:
            key_list = [k for k in dictionary.keys() if search(string_match, k)]
        except error as err:
            print(err)

    if not secondary_check:
        return {k: dictionary[k] for k in key_list}
    return {
        k: dictionary[k]
        for k in key_list
        if secondary_check[0] in dictionary[k][secondary_check[1]].keys()
    }


def handle_filepath_args(args):
    ##################################################################
    """Does filepath expansion for config args"""
    ##################################################################
    file_path_args = ["cabundle", "pwstore", "certpath", "keypath"]
    for arg in file_path_args:
        if arg in args and args[arg]:
            args[arg] = path.expanduser(args[arg])
    if (
        "connect" in args
        and args["connect"]
        and "base_directory" in args["connect"]
        and args["connect"]["base_directory"]
    ):
        args["connect"]["base_directory"] = path.expanduser(
            args["connect"]["base_directory"]
        )
    return args


def handle_boolean_args(args, argname):
    if isinstance(args[argname], str):
        return args[argname].upper() == "TRUE"
    return args[argname]


def convert_strings_to_list(args, argname):
    ##################################################################
    """convert argparsed strings to lists for an argument"""
    ##################################################################
    if argname in args:
        args[argname] = args[argname].split(",") if args[argname] else []
        return [arg.strip() for arg in args[argname] if arg.strip()]
    return []


def parse_json_arguments(args, argument):
    ##################################################################
    """Parses the json.loads arguments as dictionaries to use"""
    ##################################################################
    try:
        if argument in args and args[argument]:
            if isinstance(args[argument], dict):
                return args[argument]
            return loads(args[argument])
        return None
    except ValueError as err:
        raise JsonArgumentError(argument, err) from err


def collect_args(parsedargs):
    ##################################################################
    """Build a dict out of the argparse args Namespace object and a dict from any
    configuration files and merge the two with cli taking priority"""
    ##################################################################
    args = {
        "ignore_decrypt": False,
        "identity": getuser(),
        "cabundle": "./certs/ca-bundle",
        "keypath": "./private",
        "pwstore": "./passwords",
        "time": 10,
        "card_slot": None,
        "certpath": None,
        "escrow_users": None,
        "min_escrow": None,
        "no_cache": False,
        "noverify": None,
        "noescrow": False,
        "overwrite": False,
        "recovery": False,
        "rules": "default",
        "stdin": False,
        "theme_map": None,
        "color": True,
        "verbosity": 0,
        "SCBackend": "opensc",
        "PKCS11_module_path": "/usr/local/lib/libykcs11.dylib",
    }
    cli_args = parsedargs if isinstance(parsedargs, dict) else vars(parsedargs)
    config_args = get_config_args(cli_args["config"], cli_args)
    args.update(config_args)
    args["connect"] = parse_json_arguments(args, "connect")
    args = handle_filepath_args(args)

    fles = ["cabundle", "pwstore"]
    for key, value in cli_args.items():
        if value is not None or key not in args:
            args[key] = value
        if key in fles and not path.exists(args[key]):
            raise FileOpenError(args[key], "No such file or directory")

    # json args
    args["color"] = handle_boolean_args(args, "color")
    args["groups"] = convert_strings_to_list(args, "groups")
    args["users"] = convert_strings_to_list(args, "users")
    args["escrow_users"] = convert_strings_to_list(args, "escrow_users")
    return setup_db(args)

def build_recipient_list(args):
    ##################################################################
    """take groups and users and make a SOA for the recipients"""
    ##################################################################
    recipient_list = []
    escrow_and_recipient_list = []
    try:
        if "groups" in args and args["groups"]:
            recipient_list += parse_group_membership(args)
        if "users" in args and args["users"]:
            recipient_list += args["users"]
        recipient_list = [x.strip() for x in list(set(recipient_list))]
        escrow_and_recipient_list = (
            recipient_list + args["escrow_users"]
        )
        for user in escrow_and_recipient_list:
            if str(user) == "":
                raise NullRecipientError
        return recipient_list, escrow_and_recipient_list
    except KeyError:  # If this is a command with no users, don't worry about it
        return None

def parse_group_membership(args):
    ##################################################################
    """Concat membership of supplied groups"""
    ##################################################################
    member_list = []
    try:
        for group in args["groups"]:
            member_list += [
                user.strip()
                for user in args[group.strip()].split(",")
                if user.strip()
            ]
        return member_list
    except KeyError as err:
        raise GroupDefinitionError(str(err)) from err


def setup_db(args):
    ##################################################################
    """Setup global db engine"""
    ##################################################################
    if args["connect"] and "base_directory" in args["connect"]:
        db_path = path.join(args["connect"]["base_directory"], "rd.db")
    elif "certpath" in args and args["certpath"]:
        db_path = path.join(args["certpath"], "rd.db")
    else:
        db_path = path.join(gettempdir(), "rd.db")
    if not path.isdir(path.dirname(db_path)):
        raise FileOpenError(path.dirname(db_path), "No such file or directory")
    args["db"] = {}
    args["db"]["path"] = db_path
    args["db"]["uri"] = f"sqlite+pysqlite:///{db_path}"
    args["db"]["engine"] = create_engine(f'{args["db"]["uri"]}?check_same_thread=False')
    Base.metadata.create_all(args["db"]["engine"])
    return args


def get_config_args(config, cli_args):
    ##################################################################
    """Return the configuration from the config file"""
    ##################################################################
    try:
        with open(config, "r", encoding="ASCII") as fname:
            config_args = safe_load(fname)
        return config_args if config_args else {}
    except IOError:
        if cli_args["verbosity"] != -1:
            print("INFO: No .pkpassrc file found")
        return {}
    except (ParserError, ScannerError) as err:
        raise ConfigParseError(
            "Parsing error with config file, please check syntax"
        ) from err


def create_or_update(
    session, model, unique_identifiers=None, dont_update=None, **kwargs
):
    ###################################################################
    """Create db object if it doesn't exist, doesn't commit to db
    updates existing objects, returns the object

    session: database session to utilize,
    model: sqlalchemy model to query on
    dont_update: list of keys in kwargs that should be ignored if the record alredy exists,
        this would generally be an ID that is a permanent record id from an external data
        source
    unique_identifiers: list of the keys in 'kwargs' that identify a unique instance
    kwargs: key, value pair of columns in the database model
    """
    ###################################################################
    # Remove parameters that do not exist in the model
    # this indicates the source has updated what it sends
    if not unique_identifiers:
        # If there are not unique_identifiers listed, utilize the whole
        # object as a unique item
        query_params = kwargs
    else:
        # If unique identifiers do exit filter the kwargs down by that list
        # and then get the instance that matches that
        query_params = {k: v for k, v in kwargs.items() if k in unique_identifiers}
    instance = session.query(model).filter_by(**query_params).first()
    # If we do not have an instance yet this means that we need
    # to create a new object
    if not instance:
        session.add(model(**kwargs))
        instance = session.query(model).filter_by(**query_params).first()
    # If we do have an instance at this point, we need to ensure
    # That the new data we recieved from the data source is
    # appropriately populated in that row
    else:
        dont_update = dont_update if dont_update else []
        for key, value in kwargs.items():
            if key not in dont_update:
                setattr(instance, key, value)
    return instance



"""This Module allows for the listing of users cards"""
from libpkpass.commands.command import Command
from libpkpass.crypto import print_card_info


class Card(Command):
    ####################################################################
    """This class implements the cli card command"""
    ####################################################################

    name = "card"
    description = "List the available cards and which card you have selected"
    selected_args = Command.selected_args

    def _run_command_execution(self):
        ####################################################################
        """Run function for class."""
        ####################################################################
        return print_card_info(
            self.args["card_slot"],
            self.iddb.id,
            2,
            self.args["color"],
            self.args["theme_map"],
            self.args["SCBackend"],
        )

    def _validate_args(self):
        pass



#!/usr/bin/env python3
"""This Module handles the crypto functions i.e. encryption and decryption"""
from base64 import urlsafe_b64decode, urlsafe_b64encode
from tempfile import NamedTemporaryFile
from os import unlink, environ
from hashlib import sha256
from shutil import get_terminal_size
from subprocess import Popen, PIPE, STDOUT, DEVNULL
from cryptography.fernet import Fernet, InvalidToken
from cryptography.hazmat.backends import default_backend
from cryptography.hazmat.primitives import hashes
from cryptography.hazmat.primitives.kdf.pbkdf2 import PBKDF2HMAC
from libpkpass.util import color_prepare
from libpkpass.errors import (
    EncryptionError,
    DecryptionError,
    SignatureCreationError,
    X509CertificateError,
    BadBackendError,
    PKPassError,
)


def handle_python_strings(string):
    ####################################################################
    """handles py2/3 incompatiblities"""
    ####################################################################
    return string if isinstance(string, bytes) else string.encode("UTF-8")


def pk_encrypt_string(plaintext_string, certificate):
    ####################################################################
    """Encrypt and return a base 64 encoded string for the provided identity"""
    ####################################################################
    plaintext_derived_key = Fernet.generate_key()
    with NamedTemporaryFile(delete=False) as fname:
        fname.write(certificate)
        cert_file_path = fname.name
    command = [
        "openssl",
        "pkeyutl",
        "-encrypt",
        "-certin",
        "-inkey",
        cert_file_path,
        "-pkeyopt",
        "rsa_padding_mode:pkcs1"
    ]
    with Popen(command, stdout=PIPE, stdin=PIPE, stderr=STDOUT) as proc:
        stdout, _ = proc.communicate(input=plaintext_derived_key)
        if isinstance(certificate, bytes):
            unlink(fname.name)

        if proc.returncode != 0:
            raise EncryptionError(f"Error encrypting derived key: {stdout}")
        ciphertext_derived_key = stdout

    encrypted_string = Fernet(plaintext_derived_key).encrypt(
        handle_python_strings(plaintext_string)
    )
    return (
        encrypted_string,
        urlsafe_b64encode(handle_python_strings(ciphertext_derived_key)),
    )


def get_card_info(SCBackend="opensc"):
    if SCBackend == "opensc":
        command = ["pkcs11-tool", "-L"]
        with Popen(command, stdout=PIPE, stdin=PIPE, stderr=STDOUT) as proc:
            stdout, _ = proc.communicate()
            return (handle_python_strings(stdout).split(b"Slot"), stdout)
    elif SCBackend == "yubi":
        command = ["yubico-piv-tool", "-a", "list-readers"]
        with Popen(command, stdout=PIPE, stdin=PIPE, stderr=STDOUT) as proc:
            stdout, _ = proc.communicate()
            return (handle_python_strings(stdout).splitlines(), stdout)
    raise BadBackendError(SCBackend)


def print_card_info(card_slot, identity, verbosity, color, theme_map, SCBackend):
    ####################################################################
    """Inform the user what card is selected"""
    ####################################################################
    if "key" not in identity or not identity["key"]:
        out_list, stdout = get_card_info(SCBackend)
        if verbosity > 1:
            yield print_all_slots(stdout, color, theme_map)
        if SCBackend == "opensc":
            for out in out_list[1:]:
                stripped = out.decode("UTF-8").strip()
                if int(stripped[0]) == int(card_slot):
                    verbosity = verbosity + 1 if verbosity < 2 else 2
                    stripped = "Using Slot" + ("\n").join(stripped.split("\n")[:verbosity])
                    yield f"{color_prepare(stripped, 'info', color, theme_map)}"
        elif SCBackend == "yubi":
            for out in out_list:
                stripped = out.decode("UTF-8").strip()
                if "Yubico" not in stripped:
                    # exit(1)
                    raise PKPassError("Unsupported SC type for backend yubi.\nYubico not in:\n" + stripped)
                if int(stripped.split('CCID')[1] or 0) == int(card_slot):
                    verbosity = verbosity + 1 if verbosity < 2 else 2
                    stripped = "Using Slot" + ("\n").join(stripped.split("\n")[:verbosity])
                    yield f"{color_prepare(stripped, 'info', color, theme_map)}"
        else:
            raise BadBackendError(SCBackend)


def print_all_slots(slot_info, color, theme_map):
    ####################################################################
    """Print all slots and cards available"""
    ####################################################################
    columns = int(get_terminal_size().columns) // 4
    return (
        f'{color_prepare("#" * columns, "debug", color, theme_map)}\n'
        f'{handle_python_strings(slot_info).decode("UTF-8").strip()}\n'
        f'{color_prepare("#" * columns, "debug", color, theme_map)}\n'
    )


def pk_decrypt_string(
    ciphertext_string, ciphertext_derived_key, identity, passphrase, SCBackend="opensc", card_slot=None, PKCS11_module_path="/usr/local/lib/libykcs11.dylib"
):
    ####################################################################
    """Decrypt a base64 encoded string for the provided identity"""
    ####################################################################
    ciphertext_derived_key = handle_python_strings(ciphertext_derived_key)
    if "key" in identity and identity["key"]:
        command = ["openssl", "pkeyutl", "-decrypt", "-inkey", identity["key"], "-pkeyopt", "rsa_padding_mode:pkcs1"]
        with Popen(command, stdout=PIPE, stdin=PIPE, stderr=STDOUT) as proc:
            stdout, _ = proc.communicate(
                input=urlsafe_b64decode(ciphertext_derived_key)
            )
            returncode = proc.returncode
        plaintext_derived_key = stdout
    else:
        if SCBackend == "opensc":
            # We've got to use pkcs15-crypt for PIV cards, and it only supports pin on
            # command line (insecure) or via stdin.  So, we have to put ciphertext into
            # a file for pkcs15-crypt to read.  YUCK!
            with NamedTemporaryFile(delete=False) as fname:
                fname.write(urlsafe_b64decode(ciphertext_derived_key))
            command = [
                "pkcs15-crypt",
                "--decipher",
                "--raw",
                "--pkcs",
                "--input",
                fname.name,
            ]
            if card_slot is not None:
                command.extend(["--reader", str(card_slot)])
            command.extend(["--pin", "-"])
            with Popen(command, stdout=PIPE, stdin=PIPE, stderr=DEVNULL) as proc:
                stdout, _ = proc.communicate(input=passphrase.encode("UTF-8"))
                unlink(fname.name)
                try:
                    plaintext_derived_key = stdout
                except IndexError as err:
                    raise DecryptionError(stdout) from err
                returncode = proc.returncode
        elif SCBackend == "yubi":
            # todo: this can be improved to not use temp files
            # https://developers.yubico.com/yubico-piv-tool/YKCS11/Supported_applications/openssl_engine.html
            with NamedTemporaryFile(delete=False) as fname:
                fname.write(urlsafe_b64decode(ciphertext_derived_key))
            command = [
                "openssl",
                "pkeyutl",
                "-decrypt",
                "-engine", "pkcs11",
                "-keyform", "engine",
                "-inkey", "pkcs11:type=private;pin-value=" + passphrase + ";serial=" + get_card_serial(card_slot),
                "-pkeyopt", "rsa_padding_mode:pkcs1",
                "-in", fname.name
            ]
            with Popen(command, stdout=PIPE, stdin=PIPE, stderr=STDOUT, env=dict(environ, PKCS11_MODULE_PATH=PKCS11_module_path)) as proc:
                stdout, _ = proc.communicate(
                    input=urlsafe_b64decode(ciphertext_derived_key)
                )
                try:
                    plaintext_derived_key = str(stdout).split("\\n")[1]
                except IndexError as err:
                    raise DecryptionError(stdout) from err
                returncode = proc.returncode
        else:
            raise BadBackendError(SCBackend)

    if returncode != 0:
        raise DecryptionError(stdout)

    return (
        Fernet(plaintext_derived_key)
        .decrypt(handle_python_strings(ciphertext_string))
        .decode("UTF-8")
    )


def pk_sign_string(string, identity, passphrase, SCBackend="opensc", card_slot=None, PKCS11_module_path="/usr/local/lib/libykcs11.dylib"):
    ####################################################################
    """Compute the hash of string and create a digital signature"""
    ####################################################################
    stringhash = sha256(string.encode("UTF-8")).hexdigest()
    if "key" in identity and identity["key"]:
        command = ["openssl", "pkeyutl", "-sign", "-inkey", identity["key"]]
        with Popen(command, stdout=PIPE, stdin=PIPE, stderr=STDOUT) as proc:
            stdout, _ = proc.communicate(input=stringhash.encode("UTF-8"))
            signature = urlsafe_b64encode(handle_python_strings(stdout))
            returncode = proc.returncode
    else:
        if SCBackend == "opensc":
            # We've got to use pkcs15-crypt for PIV cards, and it only supports pin on
            # command line (insecure) or via stdin.  So, we have to put signature text
            # into a file for pkcs15-crypt to read.  YUCK!
            with NamedTemporaryFile(delete=False) as fname:
                fname.write(stringhash.encode("UTF-8"))
            with NamedTemporaryFile(delete=False) as out:
                command = [
                    "pkcs15-crypt",
                    "--sign",
                    "-i",
                    fname.name,
                    "-o",
                    out.name,
                    "--pkcs1",
                    "--raw",
                ]
                if card_slot is not None:
                    command.extend(["--reader", str(card_slot)])
                command.extend(["--pin", "-"])
                with Popen(command, stdout=PIPE, stdin=PIPE, stderr=STDOUT) as proc:
                    stdout, _ = proc.communicate(input=passphrase.encode("UTF-8"))
                returncode = proc.returncode

            with open(out.name, "rb") as sigfile:
                signature = urlsafe_b64encode(handle_python_strings(sigfile.read()))
        elif SCBackend == "yubi":
            # todo: this can be improved to not use temp files
            # https://developers.yubico.com/yubico-piv-tool/YKCS11/Supported_applications/openssl_engine.html
            with NamedTemporaryFile(delete=False) as fname:
                fname.write(stringhash.encode("UTF-8"))
            with NamedTemporaryFile(delete=False) as out:
                command = [
                    "openssl",
                    "pkeyutl",
                    "-sign",
                    "-engine", "pkcs11",
                    "-keyform", "engine",
                    "-inkey", "pkcs11:type=private;pin-value=" + passphrase + ";serial=" + get_card_serial(card_slot),
                    "-pkeyopt", "rsa_padding_mode:pkcs1",
                    "-in", fname.name,
                    "-out", out.name
                ]
                with Popen(command, stdout=PIPE, stdin=PIPE, stderr=STDOUT, env=dict(environ, PKCS11_MODULE_PATH=PKCS11_module_path)) as proc:
                    stdout, _ = proc.communicate(
                        input=stringhash.encode("UTF-8")
                    )
                returncode = proc.returncode
            with open(out.name, "rb") as sigfile:
                signature = urlsafe_b64encode(handle_python_strings(sigfile.read()))
        else:
            raise BadBackendError(SCBackend)

        unlink(fname.name)
        unlink(out.name)

    if returncode != 0:
        raise SignatureCreationError(stdout)

    return signature


def pk_verify_signature(string, signature, certs):
    ####################################################################
    """Compute the hash of string and verify the digital signature"""
    ####################################################################
    stringhash = sha256(string.encode("UTF-8")).hexdigest()
    for cert in certs:
        with NamedTemporaryFile(delete=False) as fname:
            fname.write(cert.cert_bytes)
        command = ["openssl", "pkeyutl", "-verifyrecover", "-certin", "-inkey", fname.name]
        with Popen(command, stdout=PIPE, stdin=PIPE, stderr=STDOUT) as proc:
            stdout, _ = proc.communicate(
                input=urlsafe_b64decode(handle_python_strings(signature))
            )
            unlink(fname.name)
            if stdout.decode("UTF-8") == stringhash:
                return True

    return stdout.decode("UTF-8") == stringhash


def pk_verify_chain(certificate, cabundle):
    ####################################################################
    """Verify the publickey trust chain against a CA Bundle and return True if valid"""
    ####################################################################
    command = ["openssl", "verify", "-CAfile", cabundle]
    with Popen(command, stdout=PIPE, stdin=PIPE, stderr=STDOUT) as proc:
        stdout, _ = proc.communicate(input=certificate)

    return stdout.decode("UTF-8").rstrip() == "stdin: OK"


def get_cert_fingerprint(cert):
    ####################################################################
    """Return the modulus of the x509 certificate of the identity"""
    ####################################################################
    # SHA1 Fingerprint=F9:9D:71:54:55:BE:99:24:6A:5E:E0:BB:48:F9:63:AE:A2:05:54:98
    return get_cert_element(cert, "fingerprint").split("=")[1]


def get_cert_subject(cert):
    ####################################################################
    """Return the subject DN of the x509 certificate of the identity"""
    ####################################################################
    # subject= /C=US/O=Entrust/OU=Certification Authorities/OU=Entrust Managed Services SSP CA
    return " ".join(get_cert_element(cert, "subject").split(" ")[1:])


def get_cert_issuer(cert):
    ####################################################################
    """Return the issuer DN of the x509 certificate of the identity"""
    ####################################################################
    # issuer= /C=US/O=Entrust/OU=Certification Authorities/OU=Entrust Managed Services SSP CA
    return " ".join(get_cert_element(cert, "issuer").split(" ")[1:])


def get_cert_enddate(cert):
    ####################################################################
    """Return the issuer DN of the x509 certificate of the identity"""
    ####################################################################
    return get_cert_element(cert, "enddate").split("=")[1]


def get_cert_issuerhash(cert):
    ####################################################################
    """Return the issuer DN of the x509 certificate of the identity"""
    ####################################################################
    return get_cert_element(cert, "issuer_hash")


def get_cert_subjecthash(cert):
    ####################################################################
    """Return the issuer DN of the x509 certificate of the identity"""
    ####################################################################
    return get_cert_element(cert, "subject_hash")


def get_cert_element(cert, element):
    ####################################################################
    """Return an arbitrary element of an x509 certificate"""
    ####################################################################
    command = ["openssl", "x509", "-noout", f"-{element}"]
    if element == "fingerprint":
        command.append("-sha1")
    with Popen(command, stdout=PIPE, stdin=PIPE, stderr=STDOUT) as proc:
        stdout, _ = proc.communicate(input=cert)
        if proc.returncode != 0:
            raise X509CertificateError(stdout)

    try:
        return stdout.decode("UTF-8").rstrip()
    except IndexError as err:
        raise X509CertificateError(stdout) from err


def get_card_element(element, SCBackend, card_slot=None):
    ####################################################################
    """Return an arbitrary element of a pcks15 capable device"""
    ####################################################################
    if SCBackend == "opensc":
        command = ["pkcs15-tool", "--read-certificate", "1"]
        if card_slot is not None:
            command.extend(["--reader", str(card_slot)])
    elif SCBackend == "yubi":
        command = ["yubico-piv-tool", "-a", "read-certificate", "-s", "9a"]
        if card_slot is not None and card_slot != 0:
            command.extend(["-r", str(card_slot)])
    else:
        raise BadBackendError(SCBackend)
    with Popen(command, stdout=PIPE, stdin=PIPE, stderr=STDOUT) as proc:
        stdout, _ = proc.communicate()
    if stdout.decode("utf-8").strip().lower() == "no smart card readers found.":
        raise X509CertificateError("Smartcard not detected")
    return get_cert_element(stdout, element)


def get_card_fingerprint(SCBackend, card_slot=None):
    ####################################################################
    """Return the modulus of the x509 certificate of the identity"""
    ####################################################################
    # SHA1 Fingerprint=F9:9D:71:54:55:BE:99:24:6A:5E:E0:BB:48:F9:63:AE:A2:05:54:98
    return get_card_element("fingerprint", SCBackend, card_slot=card_slot).split("=")[1]


def get_card_subject(SCBackend, card_slot=None):
    ####################################################################
    """Return the subject DN of the x509 certificate of the identity"""
    ####################################################################
    # subject= /C=US/O=Entrust/OU=Certification Authorities/OU=Entrust Managed Services SSP CA
    return " ".join(get_card_element("subject", SCBackend, card_slot=card_slot).split(" ")[1:])


def get_card_issuer(SCBackend, card_slot=None):
    ####################################################################
    """Return the issuer DN of the x509 certificate of the identity"""
    ####################################################################
    # issuer= /C=US/O=Entrust/OU=Certification Authorities/OU=Entrust Managed Services SSP CA
    return " ".join(get_card_element("issuer", SCBackend, card_slot=card_slot).split(" ")[1:])


def get_card_startdate(SCBackend, card_slot=None):
    ####################################################################
    """Return the issuer DN of the x509 certificate of the identity"""
    ####################################################################
    return get_card_element("startdate", SCBackend, card_slot=card_slot).split("=")[1]


def get_card_enddate(SCBackend, card_slot=None):
    ####################################################################
    """Return the issuer DN of the x509 certificate of the identity"""
    ####################################################################
    return get_card_element("enddate", SCBackend, card_slot=card_slot).split("=")[1]


def get_card_issuerhash(SCBackend, card_slot=None):
    ####################################################################
    """Return the issuer DN of the x509 certificate of the identity"""
    ####################################################################
    return get_card_element("issuer_hash", SCBackend, card_slot=card_slot)


def get_card_subjecthash(SCBackend, card_slot=None):
    ####################################################################
    """Return the issuer DN of the x509 certificate of the identity"""
    ####################################################################
    return get_card_element("subject_hash", SCBackend, card_slot=card_slot)


def get_card_serial(card_slot=None):
    ####################################################################
    """Return the serial element of a yubico card"""
    ####################################################################
    command = ["yubico-piv-tool", "-a", "status"]
    if card_slot is not None and card_slot != 0:
        command.extend(["-r", str(card_slot)])
    with Popen(command, stdout=PIPE, stdin=PIPE, stderr=STDOUT) as proc:
        stdout, _ = proc.communicate()
    if stdout.decode("utf-8").strip().lower() == "no smart card readers found.":
        raise X509CertificateError("Smartcard not detected")
    for line in stdout.decode("utf-8").strip().lower().splitlines():
        if "serial number:" in line:
            return line.split(":")[1].replace(" ", "").replace("\t", "")
    raise X509CertificateError("Smartcard not detected")
    return None


def sk_encrypt_string(plaintext_string, key):
    ####################################################################
    """Symmetrically Encrypt and return a base 64 encoded string using the provided secret"""
    ####################################################################
    fern = Fernet(hash_password(key))
    encrypted_string = fern.encrypt(plaintext_string.encode("UTF-8"))
    return urlsafe_b64encode(handle_python_strings(encrypted_string))


def sk_decrypt_string(ciphertext_string, key):
    ####################################################################
    """Symmetrically Decrypt a base64 encoded string using the provided key"""
    ####################################################################
    fern = Fernet(hash_password(key))
    ciphertext_string = handle_python_strings(ciphertext_string)
    try:
        plaintext_string = fern.decrypt(
            urlsafe_b64decode(handle_python_strings(ciphertext_string))
        )
        return plaintext_string.decode("UTF-8")
    except InvalidToken:
        # handle some legacy stuff from NCCS
        try:
            plaintext_string = fern.decrypt(ciphertext_string)
            return plaintext_string.decode("UTF-8")
        except InvalidToken as err:
            raise DecryptionError("Incorrect Password") from err


def hash_password(password):
    ####################################################################
    """Hash people's password"""
    ####################################################################
    password = handle_python_strings(f"{password}")
    kdf = PBKDF2HMAC(
        algorithm=hashes.SHA256(),
        length=32,
        salt=b"salt",
        iterations=100000,
        backend=default_backend(),
    )
    return urlsafe_b64encode(handle_python_strings(kdf.derive(password)))


def puppet_password(eyaml, password):
    ####################################################################
    """Give eyaml password"""
    ####################################################################
    command = [eyaml, "encrypt", "-s", password, "-o", "string"]
    with Popen(command, stdout=PIPE, stdin=PIPE, stderr=STDOUT) as proc:
        stdout, _ = proc.communicate()
    return stdout.decode().strip()



import logging
from ._version import get_versions

__version__ = get_versions()["version"]
del get_versions

LOGGER = logging.getLogger(__name__)
logging.basicConfig(
    level="DEBUG",
    format="%(message)s",
)



"""This Module defines what a passworddb should look like"""
from os import walk, path, makedirs
from multiprocessing import Manager, cpu_count, Pool
from fnmatch import fnmatch
from pylibyaml import monkey_patch_pyyaml  # pylint: disable=unused-import
from yaml import safe_load, dump
from libpkpass.password import PasswordEntry
from libpkpass.errors import PasswordIOError


class PasswordDB:
    ####################################################################
    """Password database object.  Gets and retrieves password entries from places
    passwords are stored"""
    ####################################################################

    def __init__(self, mode="Filesystem"):
        self.mode = mode
        self.pwdb = {}
        self.ignore = "*requirements.txt"

    def __repr__(self):
        return f"{self.__class__}({self.__dict__})"

    def __str__(self):
        return f"{self.__dict__}"

    def __sizeof__(self):
        return len(self.pwdb)

    def load_password_data(self, password_id, pwdb=None):
        ####################################################################
        """Load and return password from wherever it may be stored"""
        ####################################################################
        if fnmatch(password_id, self.ignore):
            return None
        if password_id not in self.pwdb.keys() and self.mode == "Filesystem":
            self.pwdb[password_id] = self.read_password_data_from_file(password_id)
        if pwdb is not None:
            pwdb[password_id] = self.pwdb[password_id]
        return self.pwdb[password_id]

    def load_from_directory(self, pwstore):
        ####################################################################
        """Load all passwords from directory"""
        ####################################################################
        with Manager() as manager:
            pool = Pool(cpu_count())
            pwdb = manager.dict()
            for fpath, _, files in walk(pwstore):
                pool.apply_async(self.parallel_loader, args=(files, fpath, pwdb))
            pool.close()
            pool.join()
            self.pwdb = dict(pwdb)

    def parallel_loader(self, files, fpath, pwdb):
        ####################################################################
        """Function to allow multiprocessing to be utilize to read passwords"""
        ####################################################################
        for passwordname in files:
            passwordpath = path.join(fpath, passwordname)
            self.load_password_data(passwordpath, pwdb)

    def save_password_data(self, password_id, overwrite=False):
        ####################################################################
        """Store a password to wherever it may be stored"""
        ####################################################################
        if self.mode == "Filesystem":
            self.write_password_data_to_file(
                self.pwdb[password_id], password_id, overwrite
            )

    def read_password_data_from_file(self, filename):
        ####################################################################
        """Open a password file, load passwords and read metadata"""
        ####################################################################
        try:
            with open(filename, "r", encoding="ASCII") as fname:
                password_data = safe_load(fname)
                password_entry = PasswordEntry()
                password_entry.metadata = password_data["metadata"]
                password_entry.recipients = password_data["recipients"]
                if "escrow" in password_data:
                    password_entry.escrow = password_data["escrow"]
            return password_entry
        except (OSError, IOError, TypeError) as err:
            raise PasswordIOError(
                f"Error reading '{filename}' perhaps a path error for the db, or malformed file"
            ) from err

    def write_password_data_to_file(self, password_data, filename, overwrite=False):
        ####################################################################
        """Write password data and metadata to the appropriate password file"""
        ####################################################################
        try:
            if not path.isdir(path.dirname(filename)):
                makedirs(path.dirname(filename))
            with open(filename, "w+", encoding="ASCII") as fname:
                passdata = {
                    key: value for key, value in password_data.todict().items() if value
                }
                fname.write(dump(passdata, default_flow_style=False))
        except (OSError, IOError) as err:
            raise PasswordIOError(f"Error creating '{filename}'") from err



"""This Module allows for the listing of recipients"""
from libpkpass import LOGGER
from libpkpass.commands.command import Command
from libpkpass.errors import CliArgumentError
from libpkpass.models.cert import Cert
from libpkpass.models.recipient import Recipient


class Listrecipients(Command):
    ####################################################################
    """This class implements the cli functionality to list recipients"""
    ####################################################################
    name = "listrecipients"
    description = "List the recipients that pkpass knows about"
    selected_args = Command.selected_args + ["stdin", "filter"]

    def _run_command_execution(self):
        ####################################################################
        """Run function for class."""
        ####################################################################
        LOGGER.info("Certificate store: %s", self.args["certpath"])
        LOGGER.info("Key store: %s", self.args["keypath"])
        LOGGER.info("CA Bundle file: %s", self.args["cabundle"])
        LOGGER.info("Looking for Key Extension: %s", self.iddb.extensions["key"])
        LOGGER.info(
            "Looking for Certificate Extension: %s",
            self.iddb.extensions["certificate"],
        )
        LOGGER.info(
            "Loaded %s identities", len(self.iddb.session.query(Recipient).all())
        )

        if "filter" in self.args and self.args["filter"]:
            identities = self.iddb.session.query(Recipient).filter(
                Recipient.name.like(self.args["filter"].replace("*", "%"))
            )
        else:
            identities = self.iddb.session.query(Recipient).all()
        for identity in identities:
            yield self._print_identity(
                identity,
                self.iddb.session.query(Cert)
                .filter(Cert.recipients.contains(identity))
                .all(),
            )

    def _print_identity(self, identity, certs):
        ####################################################################
        """Print off identity"""
        ####################################################################
        identity_list = []
        identity_list.append(f"{self.color_print(identity.name, 'first_level')}:")
        identity_list.append(f"\t{self.color_print('certs', 'second_level')}:")
        for cert in certs:
            for info in [
                "verified",
                "subject",
                "subjecthash",
                "issuer",
                "issuerhash",
                "fingerprint",
                "enddate",
            ]:
                identity_list.append(
                    f"\t\t{self.color_print(info + ':', 'third_level')} {dict(cert)[info]}"
                )
            identity_list.append("\n")
        return "\n".join(identity_list)

    def _validate_args(self):
        ####################################################################
        """Ensure arguments are appropriate for this command"""
        ####################################################################
        for argument in ["keypath"]:
            if argument not in self.args or self.args[argument] is None:
                raise CliArgumentError(f"'{argument}' is a required argument")

    def _validate_identities(self, _=None):
        ####################################################################
        """Ensure identities are appropriate for this command"""
        ####################################################################
        return



"""This module allows for the automated generation of passwords"""
from re import search
from re import error as re_error
from exrex import getone
from libpkpass.commands.command import Command
from libpkpass.errors import CliArgumentError, NotThePasswordOwnerError, RulesMapError
from libpkpass.util import parse_json_arguments


class Generate(Command):
    ####################################################################
    """This class implements the CLI functionality of automatic generation of passwords"""
    ####################################################################
    name = "generate"
    description = "Generate a new password entry and encrypt it for yourself"
    selected_args = Command.selected_args + [
        "pwname",
        "pwstore",
        "overwrite",
        "stdin",
        "keypath",
        "nopassphrase",
        "nosign",
        "card_slot",
        "escrow_users",
        "min_escrow",
        "noescrow",
        "rules",
        "rules_map",
        "description",
        "authorizer",
    ]

    def _run_command_execution(self):
        ####################################################################
        """Run function for class."""
        ####################################################################
        safe, owner = self.safety_check()

        if safe or self.args["overwrite"]:
            password = self._generate_pass()

            if "description" not in self.args or not self.args["description"]:
                self.args["description"] = input("Description: ")

            if "authorizer" not in self.args or not self.args["authorizer"]:
                self.args["authorizer"] = input("Authorizer: ")

            self.create_or_update_pass(
                password, self.args["description"], self.args["authorizer"]
            )
        else:
            raise NotThePasswordOwnerError(
                self.args["identity"], owner, self.args["pwname"]
            )
        # necessary for print statement
        yield ""

    def _generate_pass(self):
        ####################################################################
        """Generate a password based on a rules list that has been provided"""
        #######################################################################
        try:
            regex_rule = self.args["rules_map"][self.args["rules"]]
            password = getone(regex_rule)
            # The following regex search verifies it finds a match from exrex
            return search(regex_rule, password).group(0)
        except TypeError as err:
            raise RulesMapError(
                "Poorly formatted Rules Map, please check it is in json format"
            ) from err
        except re_error as err:
            raise RulesMapError("Poorly formatted regex, or unsupported") from err

    def _validate_args(self):
        self.args["rules_map"] = parse_json_arguments(self.args, "rules_map")
        for argument in ["pwname", "keypath", "rules_map"]:
            if argument not in self.args or self.args[argument] is None:
                raise CliArgumentError(f"'{argument}' is a required argument")
        if (
            self.args["rules"] not in self.args["rules_map"]
            or self.args["rules_map"][self.args["rules"]] == ""
        ):
            raise RulesMapError(f"No Rule set defined as {self.args['rules']}")



#!/usr/bin/env python3
"""This Module is the base module for custom errors to allow for reduced tracebacks and
more actionable items on the user's side"""


class PKPassError(Exception):
    """Base class for pkpass exceptions."""

    def __init__(self, arg):
        self.msg = arg


class BlankPasswordError(PKPassError):
    def __init__(self):
        self.msg = "User Provided password is blank or only spaces"


class CliArgumentError(PKPassError):
    pass


class ConfigParseError(PKPassError):
    pass


class DecryptionError(PKPassError):
    pass


class EncryptionError(PKPassError):
    pass


class BadBackendError(PKPassError):
    def __init__(self, value):
        self.msg = f"Invalid value for SCBackend: {value}"


class EscrowError(PKPassError):
    def __init__(self, field, constant, value):
        self.msg = f"{field}, must be greater than {constant}, current value: {value}"


class FileOpenError(PKPassError):
    def __init__(self, value, reason):
        self.msg = f"File {value} found in config, could not be opened due to {reason}"


class GroupDefinitionError(PKPassError):
    def __init__(self, value):
        self.msg = f"Group {value} is not defined in the config"


class JsonArgumentError(PKPassError):
    def __init__(self, value, reason):
        self.msg = f"Parse error for '{value}' because '{reason}'"


class LegacyImportFormatError(PKPassError):
    def __init__(self):
        self.msg = "Passwords in import file not in key:value notation"


class NotARecipientError(PKPassError):
    pass


class NotThePasswordOwnerError(PKPassError):
    def __init__(self, identity, owner, pwname):
        self.msg = f"User '{identity}' is not the owner of password '{pwname}', not overwriting;\n\t\
please use another password name or ask the owner ({owner}) to distribute to you"


class NullRecipientError(PKPassError):
    def __init__(self):
        self.msg = (
            "There is a blank Recipient in the list, please check for trailing commas"
        )


class PasswordIOError(PKPassError):
    pass


class PasswordMismatchError(PKPassError):
    def __init__(self):
        self.msg = "Passwords do not match"


class PasswordValidationError(PKPassError):
    def __init__(self, field, value):
        self.msg = f"Error validating password field: {field} with value {value}"


class RulesMapError(PKPassError):
    def __init__(self, reason):
        self.msg = reason


class RSAKeyError(PKPassError):
    pass


class SignatureCreationError(PKPassError):
    pass


class TrustChainVerificationError(PKPassError):
    pass


class X509CertificateError(PKPassError):
    pass


class YamlFormatError(PKPassError):
    def __init__(self, value, reason):
        self.msg = f"Error {value} due to: {reason}"



#!/usr/bin/env python3
"""This module tests the listrecipients module"""
import logging
import unittest
import pylibyaml  # pylint: disable=unused-import
import yaml
from libpkpass.commands.cli import Cli
from .basetest.basetest import patch_args


class ListrecipientsTests(unittest.TestCase):
    """This class tests the listrecipients class"""

    def setUp(self):
        logging.disable(logging.CRITICAL)
        self.out_dict = {
            "r3": {
                "certs": {
                    "verified": True,
                }
            }
        }

    def test_listrecipients(self):
        """Test a listing of our recipients"""
        self.maxDiff = None
        with patch_args(
            subparser_name="listrecipients",
            identity="r1",
            nopassphrase="true",
            filter="r3",
        ):
            out = "\n".join(Cli().run()).replace("\t", " ")
        output = yaml.safe_load(out)
        # we remove these things because the actual values depend on creation
        # moreover, some of the outputs on different operating systems appear
        # to utilize different delimiters.
        del output["r3"]["certs"]["enddate"]
        del output["r3"]["certs"]["subjecthash"]
        del output["r3"]["certs"]["issuerhash"]
        del output["r3"]["certs"]["fingerprint"]
        del output["r3"]["certs"]["subject"]
        del output["r3"]["certs"]["issuer"]
        self.assertDictEqual(output, self.out_dict)


if __name__ == "__main__":
    unittest.main()



"""This module handles the CLI for password recovery"""
from sys import stdin
from libpkpass.escrow import pk_recover_secret
from libpkpass.commands.command import Command


class Recover(Command):
    ####################################################################
    """This class implements the CLI functionality of recovery for passwords"""
    ####################################################################

    name = "recover"
    description = "Recover a password that has been distributed using escrow functions"
    selected_args = Command.selected_args + [
        "pwstore",
        "keypath",
        "nosign",
        "escrow_users",
        "min_escrow",
        "stdin",
    ]

    def _run_command_execution(self):
        ####################################################################
        """Run function for class."""
        ####################################################################
        yield "If the password returned is not correct, you may need more shares"
        if self.args["stdin"]:
            shares = "".join(stdin.readlines()).strip()
        else:
            shares = input("Enter comma separated list of shares: ")
        yield pk_recover_secret(map(str.strip, shares.split(",")))

    def _validate_args(self):
        pass

    def _validate_combinatorial_args(self):
        pass



"""This Module handles the identitydb object"""
from os import path, listdir
from datetime import datetime
from pem import parse_file
from sqlalchemy.orm import sessionmaker
from tqdm import tqdm
from libpkpass.crypto import (
    pk_verify_chain,
    get_cert_fingerprint,
    get_cert_subject,
    get_cert_issuer,
    get_cert_enddate,
    get_cert_issuerhash,
    get_cert_subjecthash,
)
from libpkpass.errors import FileOpenError, CliArgumentError
from libpkpass.models.recipient import Recipient
from libpkpass.models.cert import Cert
from libpkpass.util import create_or_update


class IdentityDB:
    ##########################################################################
    """User database class.  Contains information about the identities of and
    things pertinent to recipients and their groups and keys."""
    ##########################################################################

    def __init__(self):
        self.extensions = {"certificate": [".cert", ".crt"], "key": ".key"}
        self.cabundle = ""
        self.args = {}
        self.session = None
        self.id = None

    def __repr__(self):
        return f"{self.__class__}({self.__dict__})"

    def __str__(self):
        return f"{self.__dict__}"

    def _load_certs_from_external(self, connection_map, usedb=True):
        #######################################################################
        """Load certificates from external via use of plugin"""
        #######################################################################
        if "base_directory" in connection_map and connection_map["base_directory"]:
            del connection_map["base_directory"]
        for key, value in connection_map.items():
            connector = getattr(__import__(key.lower(), fromlist=[key]), key)(value)
            if usedb:
                with open(self.args["db"]["path"], "wb") as db_path:
                    db_path.write(connector.get_db())
            else:
                certs = connector.list_certificates()
                print("Loading certs into database")
                for name, certlist in tqdm(certs.items()):
                    self.load_db(name, certlist)
            break

    def _load_from_directory(self, fpath, filetype):
        #######################################################################
        """Helper function to read in (keys|certs) and store them correctly"""
        #######################################################################
        try:
            print("Loading certs into database")
            for fname in tqdm(listdir(fpath)):
                if fname.endswith(tuple(self.extensions[filetype])):
                    uid = fname.split(".")[0]
                    filepath = path.join(fpath, fname)
                    if filetype == "certificate":
                        self.load_db(
                            uid, certlist=[x.as_text() for x in parse_file(filepath)]
                        )
                    elif filetype == "key":
                        session = sessionmaker(bind=self.args["db"]["engine"])()
                        create_or_update(
                            session,
                            Recipient,
                            unique_identifiers=["name"],
                            **{
                                "name": uid,
                                "key": filepath,
                            },
                        )
                        session.commit()
        except OSError as error:
            raise FileOpenError(fpath, str(error.strerror)) from error

    def load_certs_from_directory(
        self, certpath, connectmap=None, nocache=False, usedb=True
    ):
        #######################################################################
        """Read in all x509 certificates from directory and name them as found"""
        #######################################################################
        session = sessionmaker(bind=self.args["db"]["engine"])()
        if not session.query(Recipient).first() or nocache:
            if connectmap:
                self._load_certs_from_external(connectmap, usedb=usedb)
            if certpath:
                self._load_from_directory(certpath, "certificate")

    def load_db(self, identity, certlist=None):
        #######################################################################
        """Read in all rsa keys from directory and name them as found"""
        #######################################################################
        try:
            session = sessionmaker(bind=self.args["db"]["engine"])()
            recipient = create_or_update(
                session, Recipient, dont_update=["key"], **{"name": identity}
            )
            for cert in certlist:
                cert = cert.encode()
                cert_dict = {}
                cert_dict["cert_bytes"] = cert
                cert_dict["verified"] = pk_verify_chain(cert, self.cabundle)
                cert_dict["fingerprint"] = get_cert_fingerprint(cert)
                cert_dict["subject"] = get_cert_subject(cert)
                cert_dict["issuer"] = get_cert_issuer(cert)
                cert_dict["enddate"] = datetime.strptime(
                    get_cert_enddate(cert), "%b %d %H:%M:%S %Y %Z"
                )
                cert_dict["issuerhash"] = get_cert_issuerhash(cert)
                cert_dict["subjecthash"] = get_cert_subjecthash(cert)
                cert = create_or_update(
                    session, Cert, unique_identifiers=["fingerprint"], **cert_dict
                )
                if cert not in recipient.certs:
                    recipient.certs.append(cert)
            session.commit()
        except KeyError as err:
            raise CliArgumentError(
                f"Error: Recipient '{identity}' is not in the recipient database"
            ) from err

    def verify_identity(self, identity):
        session = sessionmaker(bind=self.args["db"]["engine"])()
        recipient = session.query(Recipient).filter(Recipient.name == identity).first()
        if not recipient:
            return False
        for cert in (
            session.query(Cert).filter(Cert.recipients.contains(recipient)).all()
        ):
            if not cert.verified:
                return False
        return True

    def load_keys_from_directory(self, fpath):
        #######################################################################
        """Read in all rsa keys from directory and name them as found"""
        #######################################################################
        if path.isdir(fpath):
            self._load_from_directory(fpath, "key")
        # We used to print a warning on else here, but it seems unnecessary;
        # if a user is using PIVs/Smartcards, this warning could just annoy
        # them needlessly



#!/usr/bin/env python3
"""This module tests the list module"""
import unittest
import pylibyaml  # pylint: disable=unused-import
import yaml
from libpkpass.commands.cli import Cli
from libpkpass.errors import CliArgumentError
from .basetest.basetest import patch_args, ERROR_MSGS


class ListTests(unittest.TestCase):
    """This class tests the list class"""

    def setUp(self):
        self.password_list_0 = {"Passwords for 'r2'": None}
        self.password_list_1 = {
            "Passwords for 'r1'": None,
            "test/passwords/test": {"Distributor": "r1", "Name": "test"},
        }

    def test_recipient_not_in_database(self):
        """test bad recipient functionality"""
        with self.assertRaises(CliArgumentError) as context:
            with patch_args(
                subparser_name="list", identity="bleh", nopassphrase="true"
            ):
                "".join(Cli().run())
        self.assertEqual(context.exception.msg, ERROR_MSGS["rep"])

    def test_list_none(self):
        """test list functionality for no passwords"""
        with patch_args(subparser_name="list", identity="r2", nopassphrase="true"):
            out = "\n".join(Cli().run())
        output = yaml.safe_load(out)
        self.assertDictEqual(output, self.password_list_0)

    def test_list_one(self):
        """test list functionality for one password"""
        with patch_args(subparser_name="list", identity="r1", nopassphrase="true"):
            out = "\n".join(Cli().run())
        output = yaml.safe_load(out)
        self.assertDictEqual(output, self.password_list_1)


if __name__ == "__main__":
    unittest.main()



"""This module allows for the updating of passwords"""
import getpass
from os import path
from libpkpass import LOGGER
from libpkpass.util import sort
from libpkpass.password import PasswordEntry
from libpkpass.commands.command import Command
from libpkpass.errors import (
    CliArgumentError,
    PasswordMismatchError,
    NotThePasswordOwnerError,
    PasswordIOError,
    BlankPasswordError,
)
from libpkpass.models.recipient import Recipient


class Update(Command):
    ####################################################################
    """This class implements the CLI functionality of updating existing passwords"""
    ####################################################################
    name = "update"
    description = "Change a password value and redistribute to recipients"
    selected_args = Command.selected_args + [
        "pwname",
        "pwstore",
        "overwrite",
        "stdin",
        "keypath",
        "nopassphrase",
        "nosign",
        "card_slot",
        "escrow_users",
        "min_escrow",
        "noescrow",
    ]

    def _run_command_execution(self):
        ####################################################################
        """Run function for class."""
        ####################################################################
        password = PasswordEntry()
        password.read_password_data(
            path.join(self.args["pwstore"], self.args["pwname"])
        )
        safe, owner = self.safety_check()
        if safe or self.args["overwrite"]:
            self.recipient_list = list(password["recipients"].keys())
            self.recipient_list.append(str(self.args["identity"]))
            self.recipient_list = list(set(self.recipient_list))
            yield from self._confirm_recipients()
            self._validate_identities(self.recipient_list)

            password1 = getpass.getpass("Enter updated password: ")
            if password1.strip() == "":
                raise BlankPasswordError
            password2 = getpass.getpass("Enter updated password again: ")
            if password1 != password2:
                raise PasswordMismatchError

            # This is due to a poor naming convention; we don't want to go through the
            # create or update, because we are not updating a single record in the yaml
            # we are burning it to the ground and redoing the entire thing
            self.create_pass(
                password1,
                password["metadata"]["description"],
                password["metadata"]["authorizer"],
                self.recipient_list,
            )
        else:
            raise NotThePasswordOwnerError(
                self.args["identity"], owner, self.args["pwname"]
            )

    def _confirm_recipients(self):
        not_in_db = []
        in_db = [x.name for x in self.iddb.session.query(Recipient).all()]
        for recipient in self.recipient_list:
            if recipient not in in_db:
                not_in_db.append(recipient)
        if not_in_db:
            LOGGER.warning(
                "The following recipients are not in the db, removing %s",
                ", ".join(not_in_db),
            )
            self.recipient_list = [x for x in self.recipient_list if x not in not_in_db]
        yield "The following users will receive the password: "
        yield ", ".join(sort(self.recipient_list))
        correct = input("Are these correct? (y/N) ")
        if not correct or correct.lower()[0] == "n":
            self.recipient_list = input("Please enter a comma delimited list: ")
            self.recipient_list = list(
                {x.strip() for x in self.recipient_list.split(",")}
            )
            yield from self._confirm_recipients()

    def _validate_args(self):
        for argument in ["pwname", "keypath"]:
            if argument not in self.args or self.args[argument] is None:
                raise CliArgumentError(f"'{argument}' is a required argument")

    def _pre_check(self):
        if path.exists(path.join(self.args["pwstore"], self.args["pwname"])):
            return True
        raise PasswordIOError(
            f"{path.join(self.args['pwstore'], self.args['pwname'])} does not exist"
        )



"""This Modules allows for distributing created passwords to other users"""
from os import path
from tqdm import tqdm
from libpkpass import LOGGER
from libpkpass.util import dictionary_filter, sort
from libpkpass.commands.command import Command
from libpkpass.password import PasswordEntry
from libpkpass.errors import CliArgumentError
from libpkpass.models.recipient import Recipient


class Distribute(Command):
    ####################################################################
    """This Class implements the CLI functionality for ditribution"""
    ####################################################################
    name = "distribute"
    description = "Distribute existing password entry/ies to another entity [matching uses python fnmatch]"
    selected_args = Command.selected_args + [
        "pwname",
        "pwstore",
        "users",
        "groups",
        "stdin",
        "min_escrow",
        "escrow_users",
        "keypath",
        "nopassphrase",
        "nosign",
        "card_slot",
        "noescrow",
    ]

    def __init__(self, *args, **kwargs):
        Command.__init__(self, *args, **kwargs)
        self.filtered_pdb = {}

    def _run_command_execution(self):
        ####################################################################
        """Run function for class."""
        ####################################################################
        yield from self._confirm_pdb()
        self.recipient_list.append(str(self.args["identity"]))
        self.recipient_list = list(set(self.recipient_list))
        yield from self._confirm_recipients()
        for dist_pass, _ in tqdm(self.filtered_pdb.items()):
            password = PasswordEntry()
            password.read_password_data(dist_pass)
            if self.args["identity"] in password.recipients.keys():
                plaintext_pw = password.decrypt_entry(
                    self.iddb.id,
                    passphrase=self.passphrase,
                    card_slot=self.args["card_slot"],
                    SCBackend=self.args["SCBackend"],
                    PKCS11_module_path=self.args["PKCS11_module_path"],
                )
                password.add_recipients(
                    secret=plaintext_pw,
                    distributor=self.args["identity"],
                    recipients=self.recipient_list,
                    session=self.iddb.session,
                    passphrase=self.passphrase,
                    card_slot=self.args["card_slot"],
                    escrow_users=self.args["escrow_users"],
                    minimum=self.args["min_escrow"],
                    SCBackend=self.args["SCBackend"],
                    PKCS11_module_path=self.args["PKCS11_module_path"],
                )

                password.write_password_data(dist_pass)

    def _validate_args(self):
        for argument in ["pwname", "keypath"]:
            if argument not in self.args or self.args[argument] is None:
                raise CliArgumentError(f"'{argument}' is a required argument")

    def _confirm_pdb(self):
        self.filtered_pdb = dictionary_filter(
            path.join(self.args["pwstore"], self.args["pwname"]),
            self.passworddb.pwdb,
            [self.args["identity"], "recipients"],
        )
        yield "The following password files have matched:"
        yield "\n".join(self.filtered_pdb.keys())
        correct = input("Are these correct? (y/N) ")
        if not correct or correct.lower()[0] == "n":
            self.args["pwname"] = input("Please try a new filter: ")
            yield from self._confirm_pdb()

    def _confirm_recipients(self):
        not_in_db = []
        in_db = [x.name for x in self.iddb.session.query(Recipient).all()]
        for recipient in self.recipient_list:
            if recipient not in in_db:
                not_in_db.append(recipient)
        if not_in_db:
            LOGGER.warning(
                "The following recipients are not in the db, removing %s",
                ", ".join(not_in_db),
            )
            self.recipient_list = [x for x in self.recipient_list if x not in not_in_db]
        yield "The following user(s) will be added: "
        yield ", ".join(sort(self.recipient_list))
        correct = input("Are these correct? (y/N) ")
        if not correct or correct.lower()[0] == "n":
            self.recipient_list = input("Please enter a comma delimited list: ")
            self.recipient_list = list(
                {x.strip() for x in self.recipient_list.split(",")}
            )
            yield from self._confirm_recipients()



from sqlalchemy import String
from sqlalchemy.orm import relationship
from sqlalchemy.schema import Column, ForeignKey, Table
from libpkpass.models import Base

recipient_cert = Table(
    "recipient_cert",
    Base.metadata,
    Column("recipient_name", String(), ForeignKey("recipient.name")),
    Column("cert_fingerprint", String(), ForeignKey("cert.fingerprint")),
)


class Recipient(Base):
    #########################################
    """table of health status of services"""
    #########################################

    __tablename__ = "recipient"
    name = Column(String(), primary_key=True)
    key = Column(String(), nullable=True)
    certs = relationship(
        "Cert", secondary=recipient_cert, backref="recipients", lazy="joined"
    )

    def __repr__(self):
        return f"<Recipient {self.name}>"

    def __iter__(self):
        for col in self.__table__.columns:
            yield col.name, getattr(self, col.name)



#!/usr/bin/env python3
"""This Module tests iddb module"""
import unittest
import os.path
from sqlalchemy import create_engine
from sqlalchemy.orm import sessionmaker
from libpkpass.identities import IdentityDB
from libpkpass.models.recipient import Recipient


class TestBasicFunction(unittest.TestCase):
    """This class tests the iddb class"""

    def setUp(self):
        self.certdir = "test/pki/intermediate/certs"
        self.keydir = "test/pki/intermediate/private"
        self.cabundle = "test/pki/intermediate/certs/ca-bundle"
        db_path = "test/pki/intermediate/certs/rd.db"
        self.iddb = IdentityDB()
        self.iddb.args = {
            "db": {
                "uri": f"sqlite+pysqlite:///{db_path}",
                "engine": create_engine(f"sqlite+pysqlite:///{db_path}"),
            }
        }
        self.iddb.session = sessionmaker(bind=self.iddb.args["db"]["engine"])()

    def test_certificate_loading(self):
        """Load a cert into our test"""
        self.iddb.load_certs_from_directory(self.certdir, self.cabundle)
        for identity in ("r1", "r2", "r3"):
            assert identity in [
                x[0] for x in self.iddb.session.query(Recipient.name).all()
            ]

    def test_key_loading(self):
        """Load a key into our test"""
        self.iddb.load_keys_from_directory(self.keydir)
        for identity in ("r1", "r2", "r3"):
            assert identity in [
                x[0] for x in self.iddb.session.query(Recipient.name).all()
            ]
            test_id = (
                self.iddb.session.query(Recipient)
                .filter(Recipient.name == identity)
                .first()
            )
            assert os.path.isfile(test_id.key)

    def test_keycert_loading(self):
        """Load a keycert into our test"""
        self.iddb.load_certs_from_directory(self.certdir, self.cabundle)
        self.iddb.load_keys_from_directory(self.keydir)
        for identity in ("r1", "r2", "r3"):
            assert identity in [
                x[0] for x in self.iddb.session.query(Recipient.name).all()
            ]
            test_id = (
                self.iddb.session.query(Recipient)
                .filter(Recipient.name == identity)
                .first()
            )
            assert os.path.isfile(test_id.key)


if __name__ == "__main__":
    unittest.main()



"""This Module allows for the export of passwords for the purpose of importing to a new card"""
import getpass
from tqdm import tqdm
from libpkpass.commands.command import Command
from libpkpass.errors import CliArgumentError, PasswordMismatchError


class Export(Command):
    ####################################################################
    """This Class implements the cli functionality of export"""
    ####################################################################
    name = "export"
    description = "Export passwords that you have access to and encrypt with aes"
    selected_args = Command.selected_args + [
        "pwfile",
        "stdin",
        "nopassphrase",
        "card_slot",
        "nocrypto",
    ]

    def _run_command_execution(self):
        ####################################################################
        """Run function for class."""
        ####################################################################
        crypt_pass = False
        if not self.args["nocrypto"]:
            crypt_pass = getpass.getpass("Please enter a password for the encryption: ")
            verify_pass = getpass.getpass("Please enter again for verification: ")
            if crypt_pass != verify_pass:
                raise PasswordMismatchError()

        self._iterate_pdb(self.passworddb, crypt_pass)
        # necessary for print statement
        yield ""

    def _iterate_pdb(self, passworddb, crypt_pass=False):
        ####################################################################
        """Iterate through the passwords that we can decrypt"""
        ####################################################################
        uid = self.iddb.id["name"]
        all_passwords = {
            k: v for (k, v) in passworddb.pwdb.items() if uid in v.recipients.keys()
        }
        for _, password in tqdm(all_passwords.items()):
            plaintext_pw = password.decrypt_entry(
                identity=self.iddb.id,
                passphrase=self.passphrase,
                card_slot=self.args["card_slot"],
                SCBackend=self.args["SCBackend"],
                PKCS11_module_path=self.args["PKCS11_module_path"],
            )
            password.recipients[uid]["encrypted_secret"] = plaintext_pw.encode("UTF-8")
            password.write_password_data(
                self.args["pwfile"],
                overwrite=False,
                encrypted_export=not self.args["nocrypto"],
                password=crypt_pass,
                export=True,
            )

    def _validate_args(self):
        ####################################################################
        """Ensure arguments are appropriate for this command"""
        ####################################################################
        for argument in ["pwfile", "keypath"]:
            if argument not in self.args or self.args[argument] is None:
                raise CliArgumentError(f"'{argument}' is a required argument")

    def _validate_combinatorial_args(self):
        pass



#!/usr/bin/env python3
"""This module tests the create module"""
import getpass
import builtins
import unittest
import mock
from libpkpass.commands.cli import Cli
from libpkpass.errors import CliArgumentError, DecryptionError
from .basetest.basetest import patch_args, ERROR_MSGS


class CreateTests(unittest.TestCase):
    """This class tests the create class"""

    def test_create_success(self):
        """Test a success creation of a password"""
        ret = True
        try:
            with patch_args(
                subparser_name="create",
                identity="r1",
                nopassphrase="true",
                pwname="test",
                escrow_users="r2,r3,r4",
                min_escrow=2,
            ):
                with mock.patch.object(builtins, "input", lambda _: "y"):
                    with mock.patch.object(getpass, "getpass", lambda _: "y"):
                        "".join(Cli().run())
        except DecryptionError:
            ret = False
        self.assertTrue(ret)

    def test_create_no_pass(self):
        """Test what happens with pwname is not supplied"""
        with self.assertRaises(CliArgumentError) as context:
            with patch_args(
                subparser_name="create",
                identity="r1",
                nopassphrase="true",
                pwname=None,
            ):
                with mock.patch.object(builtins, "input", lambda _: "y"):
                    with mock.patch.object(getpass, "getpass", lambda _: "y"):
                        "".join(Cli().run())
        self.assertEqual(context.exception.msg, ERROR_MSGS["pwname"])


if __name__ == "__main__":
    unittest.main()



#!/usr/bin/env python3
"""This module tests the distribute module"""
import builtins
import unittest
import mock
from libpkpass.commands.cli import Cli
from libpkpass.errors import CliArgumentError, DecryptionError
from .basetest.basetest import patch_args, ERROR_MSGS


class DistributeTests(unittest.TestCase):
    """This class tests the distribute class"""

    def test_recipient_not_in_database(self):
        """Test what happens with a recipient is not within the appropriate directory"""
        with self.assertRaises(CliArgumentError) as context:
            with patch_args(
                subparser_name="distribute",
                identity="bleh",
                nopassphrase="true",
                pwname="test",
            ):
                "".join(Cli().run())
        self.assertEqual(context.exception.msg, ERROR_MSGS["rep"])

    def test_distribute_cli_error(self):
        """Test what happens with pwname is not supplied"""
        with self.assertRaises(CliArgumentError) as context:
            with patch_args(
                subparser_name="distribute",
                identity="r1",
                nopassphrase="true",
                pwname=None,
            ):
                "".join(Cli().run())
        self.assertEqual(context.exception.msg, ERROR_MSGS["pwname"])

    def test_distribute_success(self):
        """Test a successful distribute"""
        ret = True
        try:
            with patch_args(
                subparser_name="distribute",
                identity="r1",
                nopassphrase="true",
                pwname="test",
                escrow_users="r2,r3,r4",
                min_escrow=2,
            ):
                with mock.patch.object(builtins, "input", lambda _: "y"):
                    "".join(Cli().run())
        except DecryptionError:
            ret = False
        self.assertTrue(ret)


if __name__ == "__main__":
    unittest.main()



"""This Module defines what a password contains"""
from time import time
from datetime import datetime
from os import path, makedirs
from dateutil import parser
from pylibyaml import monkey_patch_pyyaml # pylint: disable=unused-import
from yaml import safe_load, safe_dump
from yaml.parser import ParserError
from yaml.scanner import ScannerError
from tqdm import tqdm
from libpkpass.escrow import pk_split_secret
from libpkpass.errors import (
    NotARecipientError,
    DecryptionError,
    PasswordIOError,
    YamlFormatError,
    X509CertificateError,
    CliArgumentError,
)
from libpkpass.crypto import (
    pk_encrypt_string,
    pk_sign_string,
    pk_decrypt_string,
    get_card_subjecthash,
    get_card_fingerprint,
    pk_verify_signature,
    sk_encrypt_string,
    get_card_startdate,
    get_card_info,
)
from libpkpass.models.recipient import Recipient
from libpkpass.models.cert import Cert


class PasswordEntry:
    ####################################################################
    """Password entry object.  Contains information about a password and
    related metadata, as well as a list of individually encrypted strings"""
    ####################################################################
    def __init__(self, **kwargs):

        self.metadata = {
            "name": None,
            "description": None,
            "creator": None,
            "authorizer": None,
            "signature": None,
            "schemaVersion": "v2",
        }

        self.metadata.update(kwargs)
        self.recipients = {}
        self.escrow = {}

    def __getitem__(self, key):
        ####################################################################
        """Grab items like with dictionaries"""
        ####################################################################
        return getattr(self, key)

    def read_escrow(self, filename):
        ####################################################################
        """Read values from escrow section"""
        ####################################################################
        try:
            with open(filename, "r", encoding="ASCII") as fname:
                password_data = safe_load(fname)
                if "escrow" in password_data.keys():
                    return password_data["escrow"]
                return {}
        except (OSError, IOError, ScannerError, ParserError):
            # we actually don't care here.. file might not exist yet
            return {}

    def process_escrow_map(
        self,
        split_secret=None,
        distributor=None,
        recipients=None,
        session=None,
        encryption_algorithm="rsautl",
        passphrase=None,
        card_slot=None,
        escrow_users=None,
        minimum=None,
        SCBackend="opensc",
        PKCS11_module_path="/usr/local/lib/libykcs11.dylib",
    ):
        ####################################################################
        """Process the escrow user map into escrow users"""
        ####################################################################
        self.escrow = {
            "metadata": {"minimum_escrow": minimum, "creator": distributor},
            "recipients": {},
        }
        i = 0
        for escrow_user in escrow_users:
            if escrow_user not in recipients:
                self.escrow["recipients"][escrow_user] = self._add_recipient(
                    escrow_user,
                    split_secret[i],
                    distributor,
                    session,
                    encryption_algorithm,
                    passphrase,
                    card_slot,
                    SCBackend,
                    PKCS11_module_path,
                )
                i += 1

    def add_escrow(self, secret=None, escrow_users=None, minimum=None):
        ####################################################################
        """Add escrow users to the recipient list of this password object"""
        ####################################################################
        return pk_split_secret(secret, escrow_users, minimum)

    def add_recipients(
        self,
        secret=None,
        distributor=None,
        recipients=None,
        session=None,
        encryption_algorithm="rsautl",
        passphrase=None,
        card_slot=None,
        escrow_users=None,
        minimum=None,
        SCBackend="opensc",
        PKCS11_module_path="/usr/local/lib/libykcs11.dylib",
    ):
        ####################################################################
        """Add recipients to the recipient list of this password object"""
        ####################################################################
        if recipients is None:
            recipients = []

        new_recipients = {
            r: self._add_recipient(
                r,
                secret=secret,
                distributor=distributor,
                session=session,
                encryption_algorithm=encryption_algorithm,
                passphrase=passphrase,
                card_slot=card_slot,
                SCBackend=SCBackend,
                PKCS11_module_path=PKCS11_module_path,
            )
            for r in tqdm(recipients, leave=False)
        }
        self.recipients.update(new_recipients)
        if escrow_users:
            # escrow_users may now be none after the set operations
            try:
                set_reduction = list((set(escrow_users) - set(recipients)))
                if (len(escrow_users) > 3) and (len(set_reduction) < 3):
                    print(
                        "warning: min_escrow requirement not met after removing password recipients from escrow user list"
                    )
                    return
                split_secret = self.add_escrow(
                    secret=secret,
                    escrow_users=escrow_users,
                    minimum=minimum,
                )
                self.process_escrow_map(
                    split_secret=split_secret,
                    distributor=distributor,
                    recipients=recipients,
                    session=session,
                    encryption_algorithm=encryption_algorithm,
                    passphrase=passphrase,
                    card_slot=card_slot,
                    escrow_users=escrow_users,
                    minimum=minimum,
                    SCBackend=SCBackend,
                    PKCS11_module_path=PKCS11_module_path,
                )
            except ValueError as err:
                print(f"Warning cannot create escrow shares, reason: {err}")
                print("Your password has been created without escrow capabilities")

    def _get_distributor(self, session, distributor, SCBackend=None):
        distributor = (
            session.query(Recipient).filter(Recipient.name == distributor).first()
        )
        try:
            distributor_hash = get_card_subjecthash(SCBackend)
        except X509CertificateError:
            distributor_hash = (
                session.query(Cert)
                .filter(Cert.recipients.contains(distributor))
                .first()
                .subjecthash
            )
        return (distributor, distributor_hash)

    def _build_encrypted_secrets(
        self, session, recipient, encryption_algorithm, secret
    ):
        encrypted_secrets = {}
        identity = session.query(Recipient).filter(Recipient.name == recipient).first()
        for cert in (
            session.query(Cert).filter(Cert.recipients.contains(identity)).all()
        ):
            if encryption_algorithm == "rsautl":
                (encrypted_secret, encrypted_derived_key) = pk_encrypt_string(
                    secret, cert.cert_bytes
                )
            encrypted_secrets[cert.fingerprint] = {
                "encrypted_secret": encrypted_secret,
                "derived_key": encrypted_derived_key,
                "recipient_hash": cert.subjecthash,
            }
        return encrypted_secrets

    def _add_recipient(
        self,
        recipient,
        secret=None,
        distributor=None,
        session=None,
        encryption_algorithm="rsautl",
        passphrase=None,
        card_slot=None,
        SCBackend="opensc",
        PKCS11_module_path="/usr/local/lib/libykcs11.dylib",
    ):
        ####################################################################
        """Add recipient or sharer to list"""
        ####################################################################
        try:
            encrypted_secrets = self._build_encrypted_secrets(
                session, recipient, encryption_algorithm, secret
            )
            distributor, distributor_hash = self._get_distributor(session, distributor, SCBackend)
            recipient_entry = {
                "encrypted_secrets": encrypted_secrets,
                "encryption_algorithm": encryption_algorithm,
                "timestamp": time(),
                "distributor": distributor.name,
                "distributor_hash": distributor_hash,
            }
            recipient_entry["signature"] = pk_sign_string(
                self._create_signable_string(recipient_entry),
                dict(distributor),
                passphrase,
                SCBackend,
                card_slot,
                PKCS11_module_path,
            )

            return recipient_entry
        except KeyError as err:
            raise NotARecipientError(
                f"Identity '{recipient}' is not on the recipient list for password '{self.metadata['name']}'"
            ) from err

    def decrypt_entry(self, identity=None, passphrase=None, card_slot=None, SCBackend=None, PKCS11_module_path="/usr/local/lib/libykcs11.dylib"):
        ####################################################################
        """Decrypt this password entry for a particular identity
        (usually the user)"""
        ####################################################################
        try:
            recipient_entry = self.recipients[identity["name"]]
        except KeyError as err:
            raise NotARecipientError(
                f"Identity '{identity}' is not on the recipient list for password '{self.metadata['name']}'"
            ) from err

        try:
            # support old yaml format
            return pk_decrypt_string(
                recipient_entry["encrypted_secret"],
                recipient_entry["derived_key"],
                identity,
                passphrase,
                SCBackend,
                card_slot,
                PKCS11_module_path,
            )
        except KeyError:
            try:
                # support rsa
                if "key" in identity and identity["key"]:
                    for _, value in recipient_entry["encrypted_secrets"].items():
                        return pk_decrypt_string(
                            value["encrypted_secret"],
                            value["derived_key"],
                            identity,
                            passphrase,
                            SCBackend,
                            card_slot,
                            PKCS11_module_path,
                        )
                else:
                    cert_key = get_card_fingerprint(SCBackend, card_slot=card_slot)
                    return pk_decrypt_string(
                        recipient_entry["encrypted_secrets"][cert_key][
                            "encrypted_secret"
                        ],
                        recipient_entry["encrypted_secrets"][cert_key]["derived_key"],
                        identity,
                        passphrase,
                        SCBackend,
                        card_slot,
                        PKCS11_module_path,
                    )
            except DecryptionError as err:
                msg = create_error_message(recipient_entry["timestamp"], card_slot, SCBackend, err)
                raise DecryptionError(
                    f"Error decrypting password named '{self.metadata['name']}'. {msg}"
                ) from err
            except KeyError as err:
                raise DecryptionError(
                    f"Error decrypting password named '{self.metadata['name']}'. Appropriate private key not found"
                ) from err
        except DecryptionError as err:
            msg = create_error_message(recipient_entry["timestamp"], card_slot, SCBackend, err)
            raise DecryptionError(
                f"Error decrypting password named '{self.metadata['name']}'. {msg}"
            ) from err

    def verify_entry(self, uid=None, iddb=None, distributor=None, certs=None):
        ####################################################################
        """Check to see if all signatures and user certificates are okay for
        the user accessing this password entry"""
        ####################################################################
        recipient_entry = self.recipients[uid].copy()
        try:
            verified = iddb.verify_identity(distributor)
        except CliArgumentError:
            return {"distributor": distributor, "sigOK": None, "certOK": None}
        signature = recipient_entry.pop("signature")
        message = self._create_signable_string(recipient_entry)
        sig_ok = pk_verify_signature(message, signature, certs)
        return {"distributor": distributor, "sigOK": sig_ok, "certOK": verified}

    def _create_signable_string(self, recipient_entry):
        ####################################################################
        """Helper function to create a string out of the information
        this PasswordEntry contains"""
        ####################################################################
        message = ""
        for key in sorted(recipient_entry.keys()):
            if key == "timestamp":
                recipient_entry[key] = f"{float(recipient_entry[key]):12.2f}"
            if key not in ["signature", "encrypted_secrets"]:
                if isinstance(recipient_entry[key], bytes):
                    recipient_entry[key] = recipient_entry[key].decode("UTF-8")
                message = message + str(key) + str(recipient_entry[key])
        return message

    def todict(self):
        ####################################################################
        """Returns a dictionary"""
        ####################################################################
        return vars(self)

    def __repr__(self):
        return f"{self.__class__}({self.__dict__})"

    def __str__(self):
        return f"{self.__dict__}"

    def read_password_data(self, filename):
        ####################################################################
        """Open a password file, load passwords and read metadata"""
        ####################################################################
        try:
            with open(filename, "r", encoding="ASCII") as fname:
                password_data = safe_load(fname)
                self.metadata = password_data["metadata"]
                self.recipients = password_data["recipients"]
                if "escrow" in password_data:
                    self.escrow = password_data["escrow"]
        except (OSError, IOError) as error:
            raise PasswordIOError(
                f"Error Opening {filename} due to {error.strerror}"
            ) from error
        except (ScannerError, ParserError) as error:
            raise YamlFormatError(str(error.problem_mark), error.problem) from error

    def write_password_data(
        self,
        filename,
        overwrite=False,
        export=False,
        encrypted_export=False,
        password=None,
    ):
        ####################################################################
        """Write password data to a password file"""
        ####################################################################
        # check if the file is in the base password directory
        incwd = path.basename(filename) == filename
        # if not ensure path exists
        try:
            if not incwd and not path.isdir(path.dirname(filename)):
                makedirs(path.dirname(filename))
            passdata = {key: value for key, value in self.todict().items() if value}
            open_method = "a" if export else "w"
            with open(filename, open_method, encoding="ASCII") as fname:
                if encrypted_export:
                    encrypted = sk_encrypt_string(
                        safe_dump(passdata, default_flow_style=False), password
                    )
                    fname.write(encrypted.decode() + "\n")
                else:
                    fname.write(safe_dump(passdata, default_flow_style=False))
        except (OSError, IOError) as error:
            raise PasswordIOError(f"Error creating '{filename}'") from error


def create_error_message(recipient_timestamp, card_slot, SCBackend="opensc", err=None):
    card_start = get_card_startdate(SCBackend)
    card_start = datetime.timestamp(parser.parse(card_start))
    distribute_time = float(recipient_timestamp)
    # Slots are indexed at 0 so when enumerating you add 1
    # For opensc there is also an additional information line so add 1 again
    if (int(card_slot) + 2 > len(get_card_info(SCBackend)[0]) and SCBackend == "opensc") or (int(card_slot) + 1 > len(get_card_info(SCBackend)[0]) and SCBackend == "yubi"):
        msg = "Attempting to use card slot that is not connected"
    elif distribute_time < card_start:
        msg = "Password distributed before this certificate was created"
    elif err is not None and "dlopen" in str(err.msg):
        msg = "Perhaps a bad path in PKCS11_module_path"
    else:
        msg = "Perhaps a bad pin/passphrase?"
    return msg



"""This module allows for opening a PkPass interactive interpreter"""
# Look, I know what I did; I'm sorry about the following file.
# I pray that there exists a better way to do this. This file
# Should be burned to the ground and never spoken of again.
# You know it, I know it, everybody knows it.

# We need to disable unused import linting because the command classes *are* actually used...
# pylint: disable=unused-import
from os import path, environ, chdir, sep
import sys
from glob import glob
from subprocess import call
from cmd import Cmd
from inspect import getmembers, isclass
from yaml.parser import ParserError
from libpkpass import LOGGER
from libpkpass.commands.command import Command
from libpkpass.errors import PKPassError, JsonArgumentError
from libpkpass.passworddb import PasswordDB
from libpkpass.util import handle_filepath_args, show_version, collect_args
from libpkpass.commands import (
    card,
    clip,
    create,
    delete,
    distribute,
    export,
    generate,
    info,
    listrecipients,
    modify,
    recover,
    rename,
    show,
    populate,
    update,
    pkinterface,
)
import libpkpass.commands.fileimport as pkimport
import libpkpass.commands.list as pklist

VERSION, *_ = show_version()


class Interpreter(Command):
    ####################################################################
    """This class implements a skeleton to call the interpreter"""
    ####################################################################

    name = "interpreter"
    description = "Interactive mode for pkpass"
    selected_args = Command.selected_args + [
        "card_slot",
        "connect",
        "escrow_users",
        "min_escrow",
        "groups",
        "keypath",
        "pwstore",
    ]

    def _run_command_execution(self):
        ####################################################################
        """Run function for class."""
        ####################################################################
        passworddb = PasswordDB()
        passworddb.load_from_directory(self.args["pwstore"])
        Interactive(self.args, passworddb).cmdloop_with_keyboard_interrupt()
        yield "Goodbye"

    def _validate_args(self):
        pass


# This is to keep argparse from sys.exiting all over the place
def pkparse_error(message):
    raise PKPassError(message)


class Interactive(Cmd, pkinterface.PkInterface):
    ####################################################################
    """This class implements the interactive interpreter functionality"""
    ####################################################################

    intro = f"""Welcome to PKPass (Public Key Based Password Manager) v{VERSION}!
Type ? to list commands"""
    prompt = "pkpass> "

    def __init__(self, args, pwdb):
        Cmd.__init__(self)
        # manually remove interpreter from command line so that argparse doesn't try to use it
        # This needs to be removed before the PkInterace init
        sys.argv.remove("interpreter")

        pkinterface.PkInterface.__init__(self)

        self.pwdb = pwdb if pwdb else PasswordDB()
        self.parser.error = pkparse_error

        # Hold onto args passed on the command line
        self.args = args
        # change our cwd so users can tab complete
        self._change_pwstore()

        # We basically need to grab the first argument and ditch it
        self.parsedargs = {}

    def cmdloop_with_keyboard_interrupt(self):
        ####################################################################
        """handle keyboard interrupts"""
        ####################################################################
        breaker = False
        first = True
        while breaker is not True:
            try:
                if first:
                    self.cmdloop()
                else:
                    self.cmdloop(intro="")
                breaker = True
            except KeyboardInterrupt:
                first = False
                sys.stdout.write("\n")

    def _load_pwdb(self):
        ####################################################################
        """reload passworddb"""
        ####################################################################
        LOGGER.info("Reloading Password Database")
        self.pwdb = PasswordDB()
        self.pwdb.load_from_directory(self.args["pwstore"])

    def _change_pwstore(self):
        ####################################################################
        """Change directory"""
        ####################################################################
        direc = self.args["pwstore"]
        if "pwstore" in self.args and direc and path.isdir(direc):
            chdir(direc)

    def _reload_config(self):
        ####################################################################
        """Change affected globals in interpreter"""
        ####################################################################
        # We still need to be able to reload other things like recipients
        # database
        config = self.args["config"]
        config_args = self.args
        try:
            config_args = collect_args({"config": config})
            if not config_args:
                config_args = self.args
        except ParserError:
            LOGGER.error("Error parsing config file")
            config_args = self.args
        finally:
            self.args = config_args
            self.args["config"] = config
            self.args = handle_filepath_args(self.args)
            self._change_pwstore()
            self._load_pwdb()

    def precmd(self, line):
        ####################################################################
        """Fix command line arguments, this is a hack to allow argparse
        function as we expect"""
        ####################################################################
        sys.argv = [sys.argv[0]]
        sys.argv.extend(line.split())
        return line

    def postcmd(self, stop, line):
        ####################################################################
        """Fix command line arguments, this is a hack to allow argparse
        function as we expect"""
        ####################################################################
        if str(line) == "edit" or "--no-cache" in line:
            self._reload_config()
        elif str(line) == "":
            pass
        elif str(line.split()[0]) in [
            "create",
            "delete",
            "import",
            "generate",
            "rename",
        ]:
            self._load_pwdb()
        return Cmd.postcmd(self, stop, line)

    def _append_slash_if_dir(self, path_arg):
        ####################################################################
        """Appending slashes for autocomplete_file_path"""
        ####################################################################
        if path_arg and path.isdir(path_arg) and path_arg[-1] != sep:
            return path_arg + sep
        return path_arg

    def autocomplete_file_path(self, _, line, begidx, endidx):
        ####################################################################
        """File path autocompletion, used with the cmd module complete_* series functions"""
        ####################################################################
        before_arg = line.rfind(" ", 0, begidx)
        if before_arg == -1:
            return None  # arg not found

        fixed = line[before_arg + 1 : begidx]  # fixed portion of the arg
        arg = line[before_arg + 1 : endidx]
        pattern = arg + "*"

        completions = []
        for fpath in glob(pattern):
            fpath = self._append_slash_if_dir(fpath)
            completions.append(fpath.replace(fixed, "", 1))
        return completions

    def default(self, line):
        ####################################################################
        """If we don't have a proper subcommand passed"""
        ####################################################################
        LOGGER.error(
            "Command '%s' not found, see help (?) for available commands", line
        )
        return False

    def do_exit(self, _):
        ####################################################################
        """Exit the application"""
        ####################################################################
        return True

    def do_version(self, _):
        ####################################################################
        """Print the version of PkPass"""
        ####################################################################
        print(VERSION)

    def do_edit(self, _):
        ####################################################################
        """Edit your configuration file, with $EDITOR"""
        ####################################################################
        try:
            call([environ["EDITOR"], self.args["config"]])
            return False
        except (IOError, SystemExit):
            return False

    def do_git(self, line):
        ####################################################################
        """If your password store is git back, you can control git from here"""
        ####################################################################
        call(["git"] + line.split())
        return False


def add_dynamic_function(module_name, class_name):
    ##############################################################################
    """This allows repetitive do_* and complete_* functions to be condensed"""
    ##############################################################################
    swap_name = module_name
    if "pk" in module_name:
        swap_name = module_name[2:]
    fn_name = "do_" + swap_name
    complete_name = "complete_" + swap_name

    def do_fn(self, _):
        ############################################
        """Use -h or --help for information"""
        ############################################
        try:
            pk_class = [
                o
                for o in getmembers(globals()[module_name], isclass)
                if str(o[0]) == class_name
            ][0][1]
            pk_class(self, pwdb=self.pwdb)
            for mesg in self.actions[swap_name].run(self.parser.parse_args()):
                if mesg:
                    print(mesg)
            return False
        except PKPassError as err:
            LOGGER.error(err)
            return False
        except SystemExit:
            return False

    def complete_fn(self, text, line, begidx, endidx):
        return self.autocomplete_file_path(text, line, begidx, endidx)

    setattr(Interactive, fn_name, do_fn)
    setattr(Interactive, complete_name, complete_fn)


for command in [
    ("card", "Card"),
    ("clip", "Clip"),
    ("create", "Create"),
    ("delete", "Delete"),
    ("distribute", "Distribute"),
    ("export", "Export"),
    ("generate", "Generate"),
    ("pkimport", "Import"),
    ("info", "Info"),
    ("pklist", "List"),
    ("listrecipients", "Listrecipients"),
    ("modify", "Modify"),
    ("recover", "Recover"),
    ("rename", "Rename"),
    ("show", "Show"),
    ("populate", "Populate"),
    ("update", "Update"),
]:
    add_dynamic_function(command[0], command[1])



"""This module is used to process the decryption of keys"""
from os import path, unlink, walk
from fnmatch import fnmatch
from tempfile import gettempdir
from libpkpass import LOGGER
from libpkpass.commands.command import Command
from libpkpass.password import PasswordEntry
from libpkpass.errors import (
    PasswordIOError,
    CliArgumentError,
    NotARecipientError,
    DecryptionError,
)
from libpkpass.models.recipient import Recipient


class Show(Command):
    ####################################################################
    """This class is used as a command object and parses information passed through
    the CLI to show passwords that have been distributed to users"""
    ####################################################################
    name = "show"
    description = "Display a password"
    selected_args = Command.selected_args + [
        "all",
        "behalf",
        "card_slot",
        "ignore_decrypt",
        "keypath",
        "nopassphrase",
        "noverify",
        "pwname",
        "pwstore",
        "recovery",
        "stdin",
    ]

    def _run_command_execution(self):
        ####################################################################
        """Run function for class."""
        ####################################################################
        password = PasswordEntry()
        if "behalf" in self.args and self.args["behalf"]:
            yield from self._behalf_prep(password)
        else:
            yield from self._show_wrapper(password)

    def _show_wrapper(self, password):
        ####################################################################
        """Wrapper for show to allow for on behalf of"""
        ####################################################################
        if self.args["all"]:
            try:
                yield from self._walk_dir(password)
            except DecryptionError as err:
                raise err
        elif self.args["pwname"] is None:
            raise PasswordIOError("No password supplied")
        else:
            yield from self._decrypt_wrapper(
                self.args["pwstore"], password, self.args["pwname"]
            )

    def _behalf_prep(self, password):
        ####################################################################
        """Create necessary temporary file for rsa key"""
        ####################################################################
        password.read_password_data(
            path.join(self.args["pwstore"], self.args["behalf"])
        )
        # allows the password to be stored outside the root of the password directory
        self.args["behalf"] = self.args["behalf"].split("/")[-1]
        temp_key = path.join(gettempdir(), f'{self.args["behalf"]}.key')
        plaintext_pw = password.decrypt_entry(
            identity=self.iddb.id,
            passphrase=self.passphrase,
            card_slot=self.args["card_slot"],
            SCBackend=self.args["SCBackend"],
            PKCS11_module_path=self.args["PKCS11_module_path"],
        )
        with open(temp_key, "w", encoding="ASCII") as fname:
            fname.write(
                "%s\n%s\n%s"
                % (
                    "-----BEGIN RSA PRIVATE KEY-----",
                    plaintext_pw.replace("-----BEGIN RSA PRIVATE KEY-----", "")
                    .replace(" -----END RSA PRIVATE KEY----", "")
                    .replace(" ", "\n")
                    .strip(),
                    "-----END RSA PRIVATE KEY-----",
                )
            )
        self.args["identity"] = self.args["behalf"]
        self.args["key_path"] = temp_key
        yield from self._show_wrapper(password)
        unlink(temp_key)

    def _walk_dir(self, password):
        ####################################################################
        """Walk our directory searching for passwords"""
        ####################################################################
        # walk returns root, dirs, and files we just need files
        for root, _, pwnames in walk(self.args['pwstore']):
            trim_root = root.replace(self.args["pwstore"], "").lstrip("/")
            for pwname in pwnames:
                if self.args["pwname"] is None or fnmatch(
                    path.join(trim_root, pwname), self.args["pwname"]
                ):
                    try:
                        yield from self._decrypt_wrapper(root, password, pwname)
                    except DecryptionError as err:
                        if self.args['ignore_decrypt']:
                            LOGGER.err(err)
                            continue
                        raise
                    except (NotARecipientError, TypeError):
                        continue

    def _handle_escrow_show(self, password):
        ####################################################################
        """This populates the user's escrow as passwords"""
        ####################################################################
        if password.escrow:
            if self.iddb.id["name"] in password.escrow["recipients"].keys():
                return password.escrow["recipients"][self.iddb.id["name"]]
        return None

    def _decrypt_wrapper(self, directory, password, pwname):
        ####################################################################
        """Decide whether to decrypt normally or for escrow"""
        ####################################################################
        if directory and password and pwname:
            password.read_password_data(path.join(directory, pwname))
            try:
                if self.args["recovery"]:
                    myescrow = self._handle_escrow_show(password)
                    if myescrow:
                        distributor = myescrow["distributor"]
                        password["recipients"][self.iddb.id["name"]] = myescrow
                        yield self._decrypt_password_entry(password, distributor)
                else:
                    distributor = password.recipients[self.iddb.id["name"]][
                        "distributor"
                    ]
                    yield self._decrypt_password_entry(password, distributor)
            except KeyError as err:
                raise NotARecipientError(
                    f"Identity '{self.iddb.id['name']}' is not on the recipient list for password '{pwname}'"
                ) from err

    def _decrypt_password_entry(self, password, distributor):
        ####################################################################
        """This decrypts a given password entry"""
        ####################################################################
        plaintext_pw = password.decrypt_entry(
            identity=self.iddb.id,
            passphrase=self.passphrase,
            card_slot=self.args["card_slot"],
            SCBackend=self.args["SCBackend"],
            PKCS11_module_path=self.args["PKCS11_module_path"],
        )
        dist_obj = (
            self.iddb.session.query(Recipient)
            .filter(Recipient.name == distributor)
            .first()
        )
        if (dist_obj and dist_obj.certs) and not self.args["noverify"]:
            result = password.verify_entry(
                self.iddb.id["name"], self.iddb, distributor, dist_obj.certs
            )
            if not result["sigOK"]:
                LOGGER.warning(
                    "Could not verify that %s correctly signed your password entry.",
                    result["distributor"],
                )
            if not result["certOK"]:
                LOGGER.warning(
                    "Could not verify the certificate authenticity of user '%s'.",
                    result["distributor"],
                )

        return f"{self.color_print(password.metadata['name'], 'first_level')}: {plaintext_pw}"

    def _validate_args(self):
        for argument in ["keypath"]:
            if argument not in self.args or self.args[argument] is None:
                raise CliArgumentError(f"'{argument}' is a required argument")

    def _pre_check(self):
        """Ensure our password exists before asking for authentication"""
        for root, _, pwnames in walk(self.args['pwstore']):
            trim_root = root.replace(self.args["pwstore"], "").lstrip("/")
            for pwname in pwnames:
                if self.args["pwname"] is None or fnmatch(
                    path.join(trim_root, pwname), self.args["pwname"]
                ):
                    return True
        raise PasswordIOError(
            f"{path.join(self.args['pwstore'], self.args['pwname'])} does not exist"
        )



#!/usr/bin/env python3
"""This module provides base testing capabilites"""
import sys
import argparse
from contextlib import contextmanager
from io import StringIO
import mock

CONFIG = "./test/.test_config"
BADPIN = "Error decrypting password named 'test'.  Perhaps a bad pin/passphrase?"
ERROR_MSGS = {
    "pwname": "'pwname' is a required argument",
    "rep": "Error: Your user 'bleh' is not in the recipient database",
}


@contextmanager
def captured_output():
    """capture stdout and stderr for use of parsing pkpass output"""
    new_out, new_err = StringIO(), StringIO()
    old_out, old_err = sys.stdout, sys.stderr
    try:
        sys.stdout, sys.stderr = new_out, new_err
        yield sys.stdout, sys.stderr
    finally:
        sys.stdout, sys.stderr = old_out, old_err


def patch_args(**kwargs):
    """Patch argparse arguments intended for use with `with` statement
    uses default config path and no color output"""
    return mock.patch(
        "argparse.ArgumentParser.parse_args",
        return_value=argparse.Namespace(config=CONFIG, color=False, **kwargs),
    )



#!/usr/bin/env python3
"""This Module tests the password entry module"""
import unittest
from sqlalchemy import create_engine
from sqlalchemy.orm import sessionmaker
from libpkpass.password import PasswordEntry
from libpkpass.identities import IdentityDB


class TestBasicFunction(unittest.TestCase):
    """This Class tests the password entry class"""

    def setUp(self):
        self.certdir = "test/pki/intermediate/certs"
        self.keydir = "test/pki/intermediate/private"
        self.cabundle = "test/pki/intermediate/certs/ca-bundle"
        self.file1 = "test/passwords/testpassword"
        self.file2 = "test/scratch/testpassword"
        self.secret = "Secret"
        self.textblob = "Testing TextField"
        self.sender = "r1"

        self.iddb = IdentityDB()
        self.iddb.id = self.sender
        self.iddb.recipient_list = ["r2", "r3"]
        db_path = "test/pki/intermediate/certs/rd.db"
        self.iddb.args = {
            "db": {
                "uri": f"sqlite+pysqlite:///{db_path}",
                "engine": create_engine(f"sqlite+pysqlite:///{db_path}"),
            },
        }
        self.iddb.session = sessionmaker(bind=self.iddb.args["db"]["engine"])()
        self.iddb.load_certs_from_directory(self.certdir, self.cabundle)
        self.iddb.load_keys_from_directory(self.keydir)

    def test_create_encrypt_decrypt(self):
        """create a password entry"""
        passwordentry = PasswordEntry(
            name="testcreate", description=self.textblob, creator="r1", authorizer="r1"
        )

        passwordentry.add_recipients(
            secret=self.secret,
            distributor="r1",
            recipients=self.iddb.recipient_list,
            session=self.iddb.session,
        )

    def test_read_write(self):
        """Read and write password entry data"""
        passwordentry = PasswordEntry()
        passwordentry.read_password_data(self.file1)
        passwordentry.write_password_data(self.file2, overwrite=True)

        with open(self.file1, "r", encoding="ASCII") as file1:
            with open(self.file2, "r", encoding="ASCII") as file2:
                self.assertTrue(file1.read() == file2.read())


if __name__ == "__main__":
    unittest.main()



"""This module allows for inspecting metadata passwords"""
from os import path, linesep
from datetime import datetime
from libpkpass.password import PasswordEntry
from libpkpass.commands.command import Command
from libpkpass.errors import CliArgumentError


class Info(Command):
    ####################################################################
    """This class implements the display of metadata to users"""
    ####################################################################
    name = "info"
    description = "Displays metadata about a password"
    selected_args = Command.selected_args + ["pwname", "pwstore"]

    def _run_command_execution(self):
        ####################################################################
        """Run function for class."""
        ####################################################################
        password = PasswordEntry()
        password.read_password_data(
            path.join(self.args["pwstore"], self.args["pwname"])
        )

        # Metadata
        yield self.color_print("Metadata:", "first_level")
        for key, value in password.metadata.items():
            yield f"  {self.color_print(key.capitalize(), 'second_level')}: {value}"

        # Escrow
        if password.escrow:
            yield self.color_print("\nEscrow Group:", "first_level")
            for key, value in password.escrow["metadata"].items():
                yield f"    {self.color_print(key.capitalize() + ':', 'second_level')} {str(value)}"

            yield f"    {self.color_print('Share Holders:', 'second_level')} {', '.join(list(password.escrow['recipients'].keys()))}"
            yield f"    {self.color_print('Total Group Share Holders:', 'second_level')} {len(list(password.escrow['recipients'].keys()))}"

            timestamp_list = [
                x["timestamp"]
                for x in list(password.escrow["recipients"].values())
                if "timestamp" in x
            ]
            if timestamp_list:
                timestamp = int(round(float(min(timestamp_list))))
                timestamp = datetime.fromtimestamp(timestamp).strftime(
                    "%Y-%m-%d %H:%M:%S"
                )
                yield f"    {self.color_print('Group creation time:', 'second_level')} {timestamp}"

        # Recipients
        yield f"{self.color_print(linesep + 'Recipients:', 'first_level')} {', '.join(list(password.recipients.keys()))}"
        yield f"{self.color_print('Total Recipients:', 'first_level')} {len(list(password.recipients.keys()))}"
        rec_timestamp = int(
            round(
                float(
                    min(
                        [
                            x["timestamp"]
                            for x in list(password.recipients.values())
                            if "timestamp" in x
                        ]
                    )
                )
            )
        )
        rec_timestamp = datetime.fromtimestamp(rec_timestamp).strftime(
            "%Y-%m-%d %H:%M:%S"
        )
        yield f"{self.color_print('Earliest distribute timestamp:', 'first_level')} {rec_timestamp}"

    def _validate_args(self):
        for argument in ["pwname", "keypath"]:
            if argument not in self.args or self.args[argument] is None:
                raise CliArgumentError(f"'{argument}' is a required argument")



#!/usr/bin/env python3
"""This module tests the update module"""
import getpass
import builtins
import unittest
import mock
from libpkpass.commands.cli import Cli
from libpkpass.errors import DecryptionError
from .basetest.basetest import patch_args


class UpdateTests(unittest.TestCase):
    """This class tests the update class"""

    def test_update_success(self):
        """Test a successful update of a password"""
        ret = True
        try:
            with patch_args(
                subparser_name="update",
                identity="r3",
                nopassphrase="true",
                pwname="gentest",
                no_cache=True,
            ):
                with mock.patch.object(builtins, "input", lambda _: "y"):
                    with mock.patch.object(getpass, "getpass", lambda _: "y"):
                        "".join(Cli().run())
        except DecryptionError:
            ret = False
        self.assertTrue(ret)



#!/usr/bin/env python3
"""This is a super class to handle commonalities between the cli and interpreter"""
from argparse import ArgumentParser
from os import path
from libpkpass.util import set_default_subparser
from libpkpass.commands import (
    card,
    clip,
    create,
    delete,
    distribute,
    export,
    generate,
    info,
    listrecipients,
    modify,
    recover,
    rename,
    show,
    populate,
    update,
    verifyinstall,
)
import libpkpass.commands.fileimport as pkimport
import libpkpass.commands.list as pklist


class PkInterface:
    ##############################################################################
    """Class for parsing command line.  Observes subclasses of Command to Register
    those commands in the actions list."""
    ##############################################################################

    def __init__(self):
        ####################################################################
        """Intialization function for class. Register all subcommands"""
        ####################################################################
        # Hash of registered subparser actions, mapping string to actual subparser
        self.actions = {}
        home = path.expanduser("~")
        defaultrc = None
        for file in [".pkpassrc", ".pkpassrc.yml", ".pkpassrc.yaml"]:
            if path.isfile(path.join(home, file)):
                defaultrc = path.join(home, file)
        if not defaultrc:
            defaultrc = path.join(home, ".pkpassrc")
        self.parser = ArgumentParser(description="Public Key Password Manager")
        self.parser.set_default_subparser = set_default_subparser
        self.parser.add_argument(
            "--config",
            type=str,
            help="Path to a PKPass configuration file.  Defaults to '~/.pkpassrc{,.yml,.yaml}'",
            default=defaultrc,
        )
        self.parser.add_argument(
            "--debug", action="store_true", help="Errors are more verbose"
        )
        self.subparsers = self.parser.add_subparsers(
            help="sub-commands", dest="subparser_name"
        )

        card.Card(self)
        clip.Clip(self)
        create.Create(self)
        delete.Delete(self)
        distribute.Distribute(self)
        export.Export(self)
        generate.Generate(self)
        pkimport.Import(self)
        info.Info(self)
        pklist.List(self)
        listrecipients.Listrecipients(self)
        modify.Modify(self)
        recover.Recover(self)
        rename.Rename(self)
        show.Show(self)
        populate.Populate(self)
        update.Update(self)
        verifyinstall.VerifyInstall(self)

    def register(self, command_obj, command_name, command_description):
        ####################################################################
        """Register command objects and names using an observer pattern"""
        ####################################################################
        self.actions[command_name] = command_obj
        parser = self.subparsers.add_parser(command_name, help=command_description)
        command_obj.register(parser)



#!/usr/bin/env python3
"""This Module implements a CLI"""
from libpkpass.util import show_version
from libpkpass.commands.pkinterface import PkInterface
from libpkpass.commands.interpreter import Interpreter


class Cli(PkInterface):
    ####################################################################
    """Class for parsing command line.  Observes subclasses of Command to Register
    those commands in the actions list."""
    ####################################################################

    def __init__(self):
        ####################################################################
        """Intialization function for class. Register all subcommands"""
        ####################################################################
        PkInterface.__init__(self)

        Interpreter(self)
        self.parser.add_argument(
            "--version", action="store_true", help="Show the version of PkPass and exit"
        )

        self.parser.set_default_subparser(self.parser, name="interpreter")
        self.parsedargs = self.parser.parse_args()

    def run(self):
        if "version" in self.parsedargs and self.parsedargs.version:
            return show_version()
        return self.actions[self.parsedargs.subparser_name].run(self.parsedargs)



#!/usr/bin/env python3
"""This module tests the generate module"""
import getpass
import builtins
import unittest
import mock
from libpkpass.commands.cli import Cli
from libpkpass.errors import CliArgumentError, DecryptionError
from .basetest.basetest import patch_args, ERROR_MSGS


class GenerateTests(unittest.TestCase):
    """This class tests the generate class"""

    def test_generate_success(self):
        """Test a successful generation of a password"""
        ret = True
        try:
            with patch_args(
                subparser_name="generate",
                identity="r3",
                nopassphrase="true",
                pwname="gentest",
            ):
                with mock.patch.object(builtins, "input", lambda _: "y"):
                    with mock.patch.object(getpass, "getpass", lambda _: "y"):
                        "".join(Cli().run())
        except DecryptionError:
            ret = False
        self.assertTrue(ret)

    def test_generate_no_pass(self):
        """test what happens with pwname is not supplied"""
        with self.assertRaises(CliArgumentError) as context:
            with patch_args(
                subparser_name="generate",
                identity="r3",
                nopassphrase="true",
                pwname=None,
            ):
                with mock.patch.object(builtins, "input", lambda _: "y"):
                    with mock.patch.object(getpass, "getpass", lambda _: "y"):
                        "".join(Cli().run())
        self.assertEqual(context.exception.msg, ERROR_MSGS["pwname"])


if __name__ == "__main__":
    unittest.main()



#!/usr/bin/env python3
# Note: to use the 'upload' functionality of this file, you must:
# $ pip install twine
import os
import shutil
from shutil import rmtree
import sys
from subprocess import Popen, PIPE, STDOUT
import getpass
from setuptools import setup, Command
import versioneer

REQUIRED = [
    "PyYAML==6.0",
    "SQLAlchemy==1.4.30",
    "colored==1.4.3",
    "cryptography==36.0.1",
    "exrex==0.10.5",
    "mock==4.0.3",
    "pem==21.2.0",
    "pylibyaml==0.1.0",
    "pyperclip==1.8.2",
    "pyseltongue==1.0.1",
    "python-dateutil==2.8.2",
    "ruamel.yaml.clib==0.2.6",
    "ruamel.yaml==0.17.20",
    "setuptools==56.0.0",
    "tqdm==4.62.3",
]

HOME = os.path.expanduser("~")
HERE = os.path.abspath(os.path.dirname(__file__))


def rcfile_check():
    rc_location = os.path.join(HOME, ".pkpassrc.yaml")
    example_location = os.path.join(HERE, "example_pkpassrc")
    if os.path.exists(rc_location):
        if os.path.isdir(rc_location):
            print("WARN: default rc file location is a directory")
        else:
            print(".pkpassrc already defined, skipping")
            return
    else:
        print("Copying example rc file to %s" % rc_location)
        shutil.copyfile(example_location, rc_location)


# Load the package's __version__.py module as a dictionary.
ABOUT = {}


class UploadCommand(Command):
    """Support setup.py upload."""

    description = "Build and publish the package."
    user_options = []

    @staticmethod
    def status(string):
        """Prints things in bold."""
        print("\033[1m{0}\033[0m".format(string))

    def initialize_options(self):
        pass

    def finalize_options(self):
        pass

    def run(self):
        try:
            self.status("Removing previous builds..")
            rmtree(os.path.join(HERE, "dist"))
        except OSError:
            pass

        self.status("Building source and wheel distribution..")
        os.system("{0} setup.py sdist bdist_while --universal".format(sys.executable))

        self.status("Uploading the package to pypi via twine..")
        os.system("twine upload dist/*")

        self.status("Pushing git tags..")
        os.system("git tag v{0}".format(ABOUT["__version__"]))
        os.system("git push --tags")

        sys.exit()


class RCFile(Command):
    """Re/Create RCFile"""

    description = "Create a pkpass rc file"
    user_options = []

    def initialize_options(self):
        pass

    def finalize_options(self):
        pass

    def user_input(self, prompt, default):
        u_input = input(prompt).strip()
        return u_input if u_input else default

    def directory_creation(self, user_prompt, default):
        directory = self.user_input(user_prompt, default) if user_prompt else default
        directory = os.path.expanduser(directory)
        if not os.path.isdir(directory):
            os.makedirs(directory)
        return str(directory)

    def file_creation(self, user_prompt, default):
        filename = self.user_input(user_prompt, default)
        filename = os.path.expanduser(filename)
        if not os.path.exists(filename):
            self.directory_creation(None, os.path.dirname(filename))
            with open(filename, "w+") as newfile:
                newfile.write("")
        return str(filename)

    def finish_run(self):
        testing = """testing versions of openssl and pkcs15-tool if version numbers return you're probably good
for sanity sake Noah's return values were:
    openssl version: LibreSSL 2.2.7
    pkcs15-tool --version: OpenSC-0.18.0, rev: eb60481f, commit-time: 2018-05-16 13:48:37 +0200
------YOUR VALUES BELOW THIS LINE -----------"""

        print(testing)
        print(
            Popen("openssl version".split(), stdout=PIPE, stdin=PIPE, stderr=STDOUT)
            .communicate()[0]
            .decode("ASCII")
        )
        print(
            Popen(
                "pkcs15-tool --version".split(), stdout=PIPE, stdin=PIPE, stderr=STDOUT
            )
            .communicate()[0]
            .decode("ASCII")
        )

    def run(self):
        args = {
            "cabundle": "",
            "card_slot": "",
            "certpath": "",
            "connect": "",
            "escrow_users": "",
            "groups": "",
            "identity": "",
            "keypath": "",
            "min_escrow": "",
            "pwstore": "",
            "time": "",
            "rules": "",
            "rules_map": "",
            "users": "",
        }
        print(
            "If not using defaults for the following paths please use full filepath or relative to home using ~"
        )
        passdb_home = os.path.join(HOME, "passdb")
        args["certpath"] = self.directory_creation(
            "Directory for certpath (defaults to ~/passdb/certs): ",
            os.path.join(passdb_home, "certs"),
        )
        args["keypath"] = self.directory_creation(
            "Directory for keypath (defaults to ~/passdb/keys): ",
            os.path.join(passdb_home, "keys"),
        )
        args["cabundle"] = self.file_creation(
            "Path to cabundle (defaults to ~/passdb/cabundles/ca.bundle): ",
            os.path.join(passdb_home, "cabundles", "ca.bundle"),
        )
        args["pwstore"] = self.directory_creation(
            "Directory for passwords (defaults to ~/passdb/passwords): ",
            os.path.join(passdb_home, "passwords"),
        )

        print(
            Popen("pkcs11-tool -L".split(), stdout=PIPE, stdin=PIPE, stderr=STDOUT)
            .communicate()[0]
            .decode("ASCII")
        )
        args["card_slot"] = self.user_input(
            "Available slots listed above, which would you like to use? (defaults to 0): ",
            "0",
        )
        args["identity"] = self.user_input(
            "What user name would you like to use? (defaults to system user): ",
            getpass.getuser(),
        )
        print(
            "Escrow users is a feature of Pkpass. Escrow allows a password to be recovered by the majority of the escrow users in the event of an emergency."
        )
        check_escrow = self.user_input("Would you like to setup escrow? ", "n")
        if check_escrow.lower()[0] == "y":
            escrow_users = self.user_input(
                "Please enter a comma seperated list of usernames: ", ""
            ).split(",")
            args["escrow_users"] = ",".join([user.strip() for user in escrow_users])
            args["min_escrow"] = self.user_input(
                "What should be the minimum number of escrow users required to unlock? ",
                len(args["escrow_users"].split(",")) - 1,
            )

        with open(os.path.join(HOME, ".pkpassrc"), "w") as fname:
            # fname.write(contents)
            for key, value in args.items():
                if value:
                    fname.write("%s: %s\n" % (key, value))
                else:
                    fname.write("#%s:\n" % key)

        self.finish_run()


class Verify(Command):
    """Verify the RC file is still valid
    use this with:
        python setup.py verify -r "/path/to/.pkpassrc" """

    description = "Lints the pkpassrc file"
    user_options = [("rcfile=", "r", "Specify the rc file path")]

    def initialize_options(self):
        self.rcfile = "~/.pkpassrc"

    def finalize_options(self):
        pass

    def check_if_recipient(self, user, certpath):
        for fname in os.listdir(certpath):
            if fname.endswith(tuple([".crt", ".cert"])):
                uid = fname.split(".")[0]
                if uid == user:
                    return True
        return False

    def check_users(self, args_dict, certpath, valid):
        args = ["escrow_users", "groups", "identity", "users"]
        for arg in args:
            users_list = []
            if arg in args_dict.keys():
                if args_dict[arg]:
                    users_list = args_dict[arg].strip().split(",")
                    for user in users_list:
                        if not self.check_if_recipient(user.strip(), certpath):
                            valid = False
                            print("'%s' found in config, not in recipientsdb" % user)
                else:
                    valid = False
                    print("'%s' found in config, but value is empty" % arg)
        return valid

    def check_paths(self, args_dict, valid):
        args = ["cabundle", "certpath", "pwstore"]
        for arg in args:
            if arg in args_dict.keys():
                if not os.path.exists(args_dict[arg]):
                    valid = False
                    print(
                        "'%s' found in config, No such file or directory: %s"
                        % (arg, args_dict[arg])
                    )
        return valid

    def run(self):
        import yaml

        valid = True
        args = [
            "cabundle",
            "card_slot",
            "certpath",
            "connect",
            "escrow_users",
            "groups",
            "identity",
            "keypath",
            "min_escrow",
            "pwstore",
            "time",
            "rules",
            "rules_map",
            "users",
            "theme_map",
            "color",
        ]
        store_args = [
            "all",
            "ignore_decrypt",
            "long_escrow",
            "noescrow",
            "nocrypto",
            "nopassphrase",
            "nosign",
            "noverify",
            "overwrite",
            "pwfile",
            "pwname",
            "recovery",
            "stdin",
        ]
        args_dict = {}
        if self.rcfile:
            with open(os.path.expanduser(self.rcfile), "r") as rcyaml:
                try:
                    args_dict = yaml.safe_load(rcyaml)
                except yaml.YAMLError as error:
                    valid = False
                    print(error)
        for arg in store_args:
            if arg in args_dict.keys():
                valid = False
                print("'%s' found in rc file, this will be ignored" % arg)

        for arg in args_dict:
            if arg not in args and arg not in store_args:
                valid = False
                print(
                    "'%s' found in rc file, does not appear to be valid argument\
(It may be a user defined group)"
                    % arg
                )

        if "certpath" in args_dict.keys():
            valid = self.check_users(args_dict, args_dict["certpath"], valid)

        valid = self.check_paths(args_dict, valid)
        if valid:
            print("Config Valid")


CMDCLASS = versioneer.get_cmdclass()
CMDCLASS["upload"] = UploadCommand
CMDCLASS["rcfile"] = RCFile
CMDCLASS["verify"] = Verify
setup(
    version=versioneer.get_version(),
    install_requires=REQUIRED,
    extras_require={"testing": ["mock", "tox"]},
    cmdclass=CMDCLASS,
)

try:
    rcfile_check()
except FileNotFoundError as err:
    print("WARN: could not copy example rc file")



"""This module allows for the creation of passwords"""
import getpass
from sys import stdin
from libpkpass.commands.command import Command
from libpkpass.errors import CliArgumentError, PasswordMismatchError, BlankPasswordError


class Create(Command):
    ####################################################################
    """This class implements the CLI functionality of creation of passwords"""
    ####################################################################
    name = "create"
    description = "Create a new password entry and encrypt it for yourself"
    selected_args = Command.selected_args + [
        "pwname",
        "pwstore",
        "overwrite",
        "stdin",
        "keypath",
        "nopassphrase",
        "nosign",
        "card_slot",
        "escrow_users",
        "min_escrow",
        "noescrow",
        "description",
        "authorizer",
    ]

    def _run_command_execution(self):
        ####################################################################
        """Run function for class."""
        ####################################################################

        if not self.args["stdin"]:
            password1 = getpass.getpass("Enter password to create: ")
            if password1.strip() == "":
                raise BlankPasswordError
            password2 = getpass.getpass("Enter password to create again: ")
            if password1 != password2:
                raise PasswordMismatchError
        else:
            password1 = stdin.read()

        if "description" not in self.args or not self.args["description"]:
            self.args["description"] = input("Description: ")

        if "authorizer" not in self.args or not self.args["authorizer"]:
            self.args["authorizer"] = input("Authorizer: ")

        self.create_or_update_pass(
            password1, self.args["description"], self.args["authorizer"]
        )
        # necessary for print statement
        yield ""

    def _validate_args(self):
        for argument in ["pwname", "keypath"]:
            if argument not in self.args or self.args[argument] is None:
                raise CliArgumentError(f"'{argument}' is a required argument")
